import json
import os.path
import zipfile

from typing import Iterable, Union, Optional

from .. import importer


class ProtonJsonImporter(importer.BaseFileImporter):
    def do_import(self, filename, **kwargs):
        # type: (str, dict) -> Iterable[Union[importer.Record, importer.SharedFolder, importer.File]]
        name, ext = os.path.splitext(filename)
        proton_export = None     # type: Optional[dict]
        if ext == '.json':
            with open(filename, 'r', encoding='utf-8') as f:
                proton_export = json.load(f)
        elif ext == '.zip':
            with zipfile.ZipFile(filename, 'r') as zf:
                proton_export = json.loads(zf.read('Proton Pass/data.json'))

        if not proton_export:
            raise Exception(f'Proton export file \"{filename}\" format is not supported')

        vaults = proton_export.get('vaults')
        if not isinstance(vaults, dict):
            return
        for vault in vaults.values():
            name = vault.get('name') or ''
            if name == 'Personal':
                name = ''
            if name:
                fol = importer.SharedFolder()
                fol.path = name
                yield fol
            items = vault.get('items')
            if not isinstance(items, list):
                continue
            for item in items:
                if not isinstance(item, dict):
                    continue
                data = item.get('data')
                if not isinstance(data, dict):
                    continue
                record_type = data.get('type') or ''
                record = importer.Record()
                if record_type == 'login':
                    record.type = 'login'
                elif record_type == 'creditCard':
                    record.type = 'bankCard'
                elif record_type == 'note':
                    record.type = 'encryptedNotes'
                else:
                    record.type = 'login'
                metadata = data.get('metadata')
                if not isinstance(metadata, dict):
                    continue
                record.title = metadata.get('name') or ''
                if not record.title:
                    continue
                record.uid = metadata.get('itemUuid') or ''
                note = metadata.get('note') or ''
                if note:
                    if record.type == 'encryptedNotes':
                        record.fields.append(importer.RecordField('note', '', note))
                    else:
                        record.notes = note
                content = data.get('content')
                if isinstance(content, dict):
                    record.login = content.get('username') or ''
                    record.password = content.get('password') or ''
                    for url in content.get('urls') or []:
                        if record.login_url:
                            record.fields.append(importer.RecordField('url', '', url))
                        else:
                            record.login_url = url
                    totp = content.get('totpUri') or ''
                    if totp:
                        record.fields.append(importer.RecordField(type=importer.FIELD_TYPE_ONE_TIME_CODE, value=totp))
                    if 'cardholderName' in content:
                        expiration = content.get('expirationDate') or ''
                        if len(expiration) == 6:
                            expiration = expiration[:2] + '/' + expiration[2:]
                        card = {
                            'cardNumber': content.get('number') or '',
                            'cardExpirationDate': expiration,
                            'cardSecurityCode': content.get('verificationNumber') or ''
                        }
                        record.fields.append(importer.RecordField('paymentCard', '', card))
                        pin = content.get('pin') or ''
                        if pin:
                            record.fields.append(importer.RecordField('pinCode', '', pin))
                extra_fields = data.get('extraFields') or []
                for extra in extra_fields:    # type: dict
                    extra_name = extra.get('fieldName')
                    extra_type = extra.get('type') or ''
                    extra_data = extra.get('data')
                    if extra_name and isinstance(extra_data, dict):
                        if extra_type in ('text', 'hidden'):
                            extra_content = extra_data.get('content') or ''
                        elif extra_type == 'totp':
                            extra_content = extra_data.get('totpUri') or ''
                        else:
                            extra_content = ''
                        if isinstance(extra_content, str) and len(extra_content) > 0:
                            if extra_type == 'text':
                                keeper_type = 'text'
                                if '\n' in extra_content:
                                    keeper_type = 'multiline'
                                record.fields.append(importer.RecordField(keeper_type, extra_name, extra_content))
                            elif extra_type == 'hidden':
                                keeper_type = 'secret'
                                record.fields.append(importer.RecordField(keeper_type, extra_name, extra_content))
                            elif extra_type == 'totp':
                                if not extra_content.startswith('otpauth:'):
                                    extra_content = f'otpauth://totp/?secret={extra_content}'
                                has_totp = any((True for x in record.fields if x.type == importer.FIELD_TYPE_ONE_TIME_CODE))
                                if has_totp:
                                    keeper_type = 'otp'
                                else:
                                    keeper_type = importer.FIELD_TYPE_ONE_TIME_CODE
                                    extra_name = ''
                                record.fields.append(importer.RecordField(keeper_type, extra_name, extra_content))

                yield record

"""
        def split_pgp_message(m):  # type: (bytes) -> Iterable[Tuple[int, bytes]]
            start_pos = 0
            while start_pos < len(m):
                tag = m[start_pos] & ~0xC0
                lb1 = m[start_pos+1]
                if lb1 < 192:
                    l = lb1
                    l_length = 1
                elif lb1 < 255:
                    lb2 = m[start_pos+2]
                    l = ((lb1 - 192) << 8) + lb2 + 192
                    l_length = 2
                else:
                    lb2 = m[start_pos+2]
                    lb3 = m[start_pos+3]
                    lb4 = m[start_pos+4]
                    lb5 = m[start_pos+5]
                    l = (lb2 << 24) + (lb3 << 16) + (lb4 << 8) + lb5
                    l_length = 5

                start_pos += l + 1 + l_length
                data = m[start_pos-l:start_pos]
                yield tag, data
                
        no = 0
        for t, d in split_message(ee):
            no += 1
            if t == 3:
                self.assertEqual(d[0], 4)
                alg = d[1]
                self.assertEqual(d[2], 3)
                hash_alg = d[3]
                salt = d[4:12]
                count = d[12]
                hashable_length = (16 + (count & 0x0f)) << ((count >> 4) + 6)
                s2k_element = salt + 'password'
                l = len(s2k_element)
                while (s2k_element) > l:
                    // hash s2k_element
                    s2k_element -= l
                if len(d) > 12:
                    sess_alg = d[13]
                    session_key = d[14:46]
            elif t == 18:
                pass
                
"""


#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2025 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

"""
Centralized error handling utilities for biometric authentication
"""
import logging

from ...error import CommandError
from .constants import ERROR_MESSAGES

class BiometricErrorHandler:
    """Centralized error handling for biometric operations"""
    
    @staticmethod
    def handle_authentication_error(error: Exception, platform_name: str = "Biometric") -> Exception:
        """Handle authentication errors with consistent messaging"""
        error_msg = str(error).lower()
        error_str = str(error)
        
        if any(keyword in error_msg for keyword in ["cancelled", "denied"]):
            return Exception(f"{platform_name} authentication cancelled")
        elif "timeout" in error_msg:
            return Exception(f"{platform_name} authentication timed out")
        elif "not available" in error_msg:
            return Exception(f"{platform_name} is not available or not set up")
        elif any(phrase in error_msg for phrase in ["no identities are enrolled", "biometry is not enrolled", "not enrolled", "biometric not set up", "touch id not set up"]):
            return Exception(f"{platform_name} is not set up - please enroll your biometric credentials in system settings")
        elif "parameter is incorrect" in error_msg:
            return Exception(f"{platform_name} parameter error - please check your biometric setup")
        elif any(phrase in error_msg for phrase in ["not configured", "not enabled", "unavailable", "not supported"]):
            return Exception(f"{platform_name} is not available or not configured")
        elif "no matching credential found" in error_msg:
            return Exception(ERROR_MESSAGES['no_matching_credential'])
        else:
            if "error domain=" in error_msg or "code=" in error_msg:
                if "localizeddescription=" in error_msg:
                    desc_start = error_msg.find("localizeddescription=") + len("localizeddescription=")
                    desc_end = error_msg.find("}", desc_start)
                    if desc_end == -1:
                        desc_end = len(error_msg)
                    description = error_str[desc_start:desc_end].strip()
                    return Exception(f"{platform_name} error: {description}")
            
            return Exception(f"{platform_name} authentication failed: {error_str}")

    @staticmethod
    def handle_credential_creation_error(error: Exception, platform_name: str = "Biometric") -> Exception:
        """Handle credential creation errors with consistent messaging"""
        error_msg = str(error).lower()
        error_str = str(error)
        
        if any(keyword in error_msg for keyword in ["cancelled", "denied"]):
            return Exception(f"{platform_name} registration cancelled")
        elif ("object already exists" in error_msg or 
            ("oserror" in error_msg and "22" in error_msg and "object already exists" in error_msg)):
            return Exception(f"A biometric credential for this account already exists")
        elif "timeout" in error_msg:
            return Exception(f"{platform_name} registration timed out")
        elif "not available" in error_msg:
            return Exception(f"{platform_name} is not available or not set up")
        elif any(phrase in error_msg for phrase in ["no identities are enrolled", "biometry is not enrolled", "not enrolled", "biometric not set up", "touch id not set up"]):
            return Exception(f"{platform_name} is not set up - please enroll your biometric credentials in system settings")
        elif any(phrase in error_msg for phrase in ["not configured", "not enabled", "unavailable", "not supported"]):
            return Exception(f"{platform_name} is not available or not configured")
        else:
            if "error domain=" in error_msg or "code=" in error_msg:
                if "localizeddescription=" in error_msg:
                    desc_start = error_msg.find("localizeddescription=") + len("localizeddescription=")
                    desc_end = error_msg.find("}", desc_start)
                    if desc_end == -1:
                        desc_end = len(error_msg)
                    description = error_str[desc_start:desc_end].strip()
                    return Exception(f"{platform_name} error: {description}")
            
            return Exception(f"{platform_name} registration failed: {error_str}")

    @staticmethod
    def handle_command_error(command_name: str, operation: str, error: Exception):
        """Handle command errors with clean error messages"""
        raise CommandError(None, str(error))

    @staticmethod
    def handle_keyboard_interrupt(command_name: str, operation: str):
        """Handle keyboard interrupt with clean error messages"""
        raise CommandError(None, f'{operation} cancelled by user')

    @staticmethod
    def create_platform_error(platform_name: str, error_key: str, additional_info: str = "") -> str:
        """Create platform-specific error message"""
        base_message = ERROR_MESSAGES.get(error_key, f"Unknown error in {platform_name}")
        if additional_info:
            return f"{base_message}: {additional_info}"
        return base_message

    @staticmethod
    def validate_dependencies(required_modules: list, platform_name: str = "Platform") -> None:
        """Validate that required modules are available"""
        missing_modules = []
        for module in required_modules:
            try:
                __import__(module)
            except ImportError:
                missing_modules.append(module)
        
        if missing_modules:
            raise Exception(f"Required {platform_name} dependencies not available: {', '.join(missing_modules)}")

    @staticmethod
    def create_storage_error(operation: str, platform_name: str, error: Exception) -> None:
        """Log storage operation errors consistently"""
        logging.debug(f'Failed to {operation} {platform_name} biometric flag: {error}')

    @staticmethod
    def execute_with_error_handling(command_name: str, operation: str, func, *args, **kwargs):
        """Execute a function with consistent error handling"""
        try:
            return func(*args, **kwargs)
        except KeyboardInterrupt:
            BiometricErrorHandler.handle_keyboard_interrupt(command_name, operation)
        except Exception as e:
            BiometricErrorHandler.handle_command_error(command_name, operation, e) 


from __future__ import annotations
import argparse
import json
import logging
from . import PAMGatewayActionDiscoverCommandBase, GatewayContext
from ..pam.router_helper import router_get_connected_gateways
from ... import vault_extensions
from ...display import bcolors
from ...discovery_common.jobs import Jobs
from ...discovery_common.infrastructure import Infrastructure
from ...discovery_common.constants import DIS_INFRA_GRAPH_ID
from ...discovery_common.types import DiscoveryDelta, DiscoveryObject
from ...keeper_dag.dag import DAG
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ...discovery_common.jobs import JobItem


def _h(text):
    return f"{bcolors.HEADER}{text}{bcolors.ENDC}"


def _f(text):
    return f"{bcolors.FAIL}{text}{bcolors.ENDC}"


def _g(text):
    return f"{bcolors.OKGREEN}{text}{bcolors.ENDC}"


def _b(text):
    return f"{bcolors.OKBLUE}{text}{bcolors.ENDC}"


class PAMGatewayActionDiscoverJobStatusCommand(PAMGatewayActionDiscoverCommandBase):
    parser = argparse.ArgumentParser(prog='pam-action-discover-status')
    parser.add_argument('--gateway', '-g', required=False, dest='gateway', action='store',
                        help='Show only discovery jobs from a specific gateway.')
    parser.add_argument('--job-id', '-j', required=False, dest='job_id', action='store',
                        help='Detailed information for a specific discovery job.')
    # parser.add_argument('--file', required=False, dest='json_file', action='store',
    #                     help='Save status to JSON file.')
    parser.add_argument('--history', required=False, dest='show_history', action='store_true',
                        help='Show history')

    def get_parser(self):
        return PAMGatewayActionDiscoverJobStatusCommand.parser

    def job_detail(self, job):
        pass

    @staticmethod
    def print_job_table(jobs, max_gateway_name, show_history=False):

        print("")
        print(f"{bcolors.HEADER}{'Job ID'.ljust(14, ' ')} "
              f"{'Gateway Name'.ljust(max_gateway_name, ' ')} "
              f"{'Gateway UID'.ljust(22, ' ')} "
              f"{'Status'.ljust(12, ' ')} "
              f"{'Resource UID'.ljust(22, ' ')} "
              f"{'Started'.ljust(19, ' ')} "
              f"{'Completed'.ljust(19, ' ')} "
              f"{'Duration'.ljust(19, ' ')} "
              f"{bcolors.ENDC}")

        print(f"{''.ljust(14, '=')} "
              f"{''.ljust(max_gateway_name, '=')} "
              f"{''.ljust(22, '=')} "
              f"{''.ljust(12, '=')} "
              f"{''.ljust(22, '=')} "
              f"{''.ljust(19, '=')} "
              f"{''.ljust(19, '=')} "
              f"{''.ljust(19, '=')}")

        completed_jobs = []
        running_jobs = []
        failed_jobs = []

        for job in jobs:
            color = ""
            job_id = job['job_id']
            if job['status'] == "COMPLETE":
                color = bcolors.OKGREEN
                completed_jobs.append(job_id)
            elif job['status'] == "RUNNING":
                color = bcolors.OKBLUE
                running_jobs.append(job_id)
            elif job['status'] == "FAILED":
                failed_jobs.append(job_id)
                color = bcolors.FAIL
            print(f"{color}{job_id} "
                  f"{job['gateway'].ljust(max_gateway_name, ' ')} "
                  f"{job['gateway_uid']} "
                  f"{job['status'].ljust(12, ' ')} "
                  f"{(job.get('resource_uid') or 'NA').ljust(22, ' ')} "
                  f"{(job.get('start_ts_str') or 'NA').ljust(19, ' ')} "
                  f"{(job.get('end_ts_str') or 'NA').ljust(19, ' ')} "
                  f"{(job.get('duration') or 'NA').ljust(19, ' ')} "
                  f"{bcolors.ENDC}")

        if len(completed_jobs) > 0 and show_history is False:
            print("")
            if len(completed_jobs) == 1:
                print(f"There is one {_g('COMPLETED')} job. To process, use the following command.")
            else:
                print(f"There are {len(completed_jobs)} {_g('COMPLETED')} jobs. "
                      "To process, use one of the the following commands.")
            for job_id in completed_jobs:
                print(_g(f"  pam action discover process -j {job_id}"))

        if len(running_jobs) > 0 and show_history is False:
            print("")
            if len(running_jobs) == 1:
                print(f"There is one {_b('RUNNING')} job. "
                      "If there is a problem, use the following command to cancel/remove the job.")
            else:
                print(f"There are {len(running_jobs)} {_b('RUNNING')} jobs. "
                      "If there is a problem, use one of the following commands to cancel/remove the job.")
            for job_id in running_jobs:
                print(_b(f"  pam action discover remove -j {job_id}"))

        if len(failed_jobs) > 0 and show_history is False:
            print("")
            if len(failed_jobs) == 1:
                print(f"There is one {_f('FAILED')} job. "
                      "If there is a problem, use the following command to get more information.")
            else:
                print(f"There are {len(failed_jobs)} {_f('FAILED')} jobs. "
                      "If there is a problem, use one of the following commands to get more information.")
            for job_id in failed_jobs:
                print(_f(f"  pam action discover status -j {job_id}"))
            print("")
            if len(failed_jobs) == 1:
                print(f"To remove the job, use the following command.")
            else:
                print(f"To remove the {_f('FAILED')} job, use one of the following commands.")
            for job_id in failed_jobs:
                print(_f(f"  pam action discover remove -j {job_id}"))

        print("")

    @staticmethod
    def print_job_detail(params, gateway_context, jobs, job_id):

        infra = Infrastructure(record=gateway_context.configuration, params=params)

        for job in jobs:
            if job_id == job["job_id"]:
                gateway_context = job["gateway_context"]
                if job['status'] == "COMPLETE":
                    color = bcolors.OKGREEN
                elif job['status'] == "RUNNING":
                    color = bcolors.OKBLUE
                else:
                    color = bcolors.FAIL
                status = f"{color}{job['status']}{bcolors.ENDC}"

                print("")
                print(f"{_h('Job ID')}: {job['job_id']}")
                print(f"{_h('Sync Point')}: {job['sync_point']}")
                print(f"{_h('Gateway Name')}: {job['gateway']}")
                print(f"{_h('Gateway UID')}: {job['gateway_uid']}")
                print(f"{_h('Configuration UID')}: {gateway_context.configuration_uid}")
                print(f"{_h('Status')}: {status}")
                print(f"{_h('Resource UID')}: {job.get('resource_uid', 'NA')}")
                print(f"{_h('Started')}: {job['start_ts_str']}")
                print(f"{_h('Completed')}: {job.get('end_ts_str')}")
                print(f"{_h('Duration')}: {job.get('duration')}")

                # If it failed, show the error and stacktrace.
                if job['status'] == "FAILED":
                    print("")
                    print(f"{_h('Gateway Error')}:")
                    print(f"{color}{job['error']}{bcolors.ENDC}")
                    print("")
                    print(f"{_h('Gateway Stacktrace')}:")
                    print(f"{color}{job['stacktrace']}{bcolors.ENDC}")
                # If it finished, show information about what was discovered.
                elif job.get('end_ts') is not None:
                    job_item = job.get("job_item")   # type: JobItem

                    try:
                        infra.load(sync_point=0)
                        print("")
                        delta_json = job.get('delta')
                        if delta_json is not None:
                            delta = DiscoveryDelta.model_validate(delta_json)
                            print(f"{_h('Added')} - {len(delta.added)} count")
                            for item in delta.added:
                                vertex = infra.dag.get_vertex(item.uid)
                                if vertex is None or vertex.active is False or vertex.has_data is False:
                                    logging.debug("added: vertex is none, inactive or has no data")
                                    continue
                                discovery_object = DiscoveryObject.get_discovery_object(vertex)
                                print(f"  * {discovery_object.description}")

                            print("")
                            print(f"{_h('Changed')} - {len(delta.changed)} count")
                            for item in delta.changed:
                                vertex = infra.dag.get_vertex(item.uid)
                                if vertex is None or vertex.active is False or vertex.has_data is False:
                                    logging.debug("changed: vertex is none, inactive or has no data")
                                    continue
                                discovery_object = DiscoveryObject.get_discovery_object(vertex)
                                print(f"  * {discovery_object.description}")
                                if item.changes is None:
                                    print(f"    no changed, may be a object not added in prior discoveries.")
                                else:
                                    for key, value in item.changes.items():
                                        print(f"    - {key} = {value}")

                            print("")
                            print(f"{_h('Deleted')} - {len(delta.deleted)} count")
                            for item in delta.deleted:
                                print(f"  * discovery vertex {item.uid}")
                        else:
                            print(f"{_f('There are no available delta changes for this job.')}")

                    except Exception as err:
                        print(f"{_f('Could not load delta from infrastructure: ' + str(err))}")
                        print("Fall back to raw graph.")
                        print("")
                        dag = DAG(conn=infra.conn, record=infra.record, graph_id=DIS_INFRA_GRAPH_ID)
                        print(dag.to_dot_raw(sync_point=job_item.sync_point, rank_dir="RL"))

                return

        print(f"{bcolors.FAIL}Cannot find the job{bcolors.ENDC}")

    def execute(self, params, **kwargs):

        if not hasattr(params, 'pam_controllers'):
            router_get_connected_gateways(params)

        gateway_filter = kwargs.get("gateway")
        job_id = kwargs.get("job_id")
        show_history = kwargs.get("show_history")

        if job_id is not None:
            show_history = True

        # Get all the PAM configuration records
        configuration_records = list(vault_extensions.find_records(params, "pam.*Configuration"))

        # This is used to format the table. Start with a length of 12 characters for the gateway.
        max_gateway_name = 12

        all_jobs = []
        all_gateways = GatewayContext.all_gateways(params)

        # For each configuration/ gateway, we are going to get all jobs.
        # We are going to query the gateway for any updated status.
        gateway_context = None
        for configuration_record in configuration_records:

            gateway_context = GatewayContext.from_configuration_uid(params=params,
                                                                    configuration_uid=configuration_record.record_uid,
                                                                    gateways=all_gateways)
            if gateway_context is None:
                continue

            # If we are using a gateway filter, and this gateway is not the one, then go onto the next conf/gateway.
            if gateway_filter is not None and gateway_context.is_gateway(gateway_filter) is False:
                continue

            # If the gateway name is longer that the prior, set the max length to this gateway's name.
            if len(gateway_context.gateway_name) > max_gateway_name:
                max_gateway_name = len(gateway_context.gateway_name)

            jobs = Jobs(record=configuration_record, params=params)
            if show_history is True:
                job_list = reversed(jobs.history)
            else:
                job_list = []
                if jobs.current_job is not None:
                    job_list = [jobs.current_job]

            for job_item in job_list:
                job = job_item.model_dump()
                job["status"] = "RUNNING"
                if job_item.start_ts is not None:
                    job["start_ts_str"] = job_item.start_ts_str
                if job_item.end_ts is not None:
                    job["end_ts_str"] = job_item.end_ts_str
                    job["status"] = "COMPLETE"

                job["duration"] = job_item.duration_sec_str

                job["gateway"] = gateway_context.gateway_name
                job["gateway_uid"] = gateway_context.gateway_uid

                # This is needs for details
                job["gateway_context"] = gateway_context
                job["job_item"] = job_item

                if job_item.success is False:
                    job["status"] = "FAILED"

                all_jobs.append(job)

        # Instead of printing a table, save a json file.
        if kwargs.get("json_file") is not None:
            with open(kwargs.get("json_file"), "w") as fh:
                fh.write(json.dumps(all_jobs, indent=4))
                fh.close()
            return

        if len(all_jobs) == 0:
            print(f"{bcolors.FAIL}There are no discovery jobs. Use 'pam action discover start' to start a "
                  f"discovery job.{bcolors.ENDC}")
            return

        if job_id is not None and gateway_context is not None:
            self.print_job_detail(params, gateway_context, all_jobs, job_id)
        else:
            self.print_job_table(all_jobs, max_gateway_name, show_history)



def value_to_boolean(value):
    value = str(value)
    if value.lower() in ['true', 'yes', 'on', '1']:
        return True
    elif value.lower() in ['false', 'no', 'off', '0']:
        return False
    else:
        return None



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2025 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import argparse

from .base import BiometricCommand
from ..utils.aaguid import get_provider_name_from_aaguid


class BiometricListCommand(BiometricCommand):
    """List biometric authentication methods"""

    parser = argparse.ArgumentParser(prog='biometric list', description='List biometric authentication methods')

    def get_parser(self):
        return self.parser

    def execute(self, params, **kwargs):
        """List registered biometric methods"""
        def _list():
            passkeys = self.client.get_available_credentials(params)
            self._display_credentials(passkeys)

        return self._execute_with_error_handling('list biometric methods', _list)

    def _display_credentials(self, passkeys):
        """Display credentials in table format"""
        if not passkeys:
            print("No biometric authentication methods found.")
        else:
            print("\nRegistered Biometric Authentication Methods:")
            print("-" * 70)
            for passkey in passkeys:
                created_date = self._format_timestamp(passkey.get('created'))
                last_used_date = self._format_timestamp(passkey.get('last_used'))
                
                display_name = passkey['name']
                # If name is empty, use provider name from AAGUID
                if not display_name:
                    aaguid = passkey.get('aaguid')
                    provider_name = get_provider_name_from_aaguid(aaguid)
                    display_name = provider_name
                
                print(f"Name: {display_name}")
                print(f"Created: {created_date}")
                print(f"Last Used: {last_used_date}")
                print("-" * 70) 


#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2025 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import logging
import subprocess
import base64
from typing import Optional

from .... import crypto
from ...utils.constants import MACOS_KEYCHAIN_SERVICE_PREFIX, DEFAULT_BIOMETRIC_TIMEOUT
from ...utils.error_handler import BiometricErrorHandler


class MacOSKeychainManager:
    """macOS Keychain credential manager"""
    
    def __init__(self):
        self.service_prefix = MACOS_KEYCHAIN_SERVICE_PREFIX
    
    def _authenticate_with_touchid(self, service_name: str, account_name: str, timeout_seconds: Optional[float] = None) -> Optional[str]:
        """Authenticate with Touch ID and return the credential data"""
        try:
            timeout = timeout_seconds or DEFAULT_BIOMETRIC_TIMEOUT
            return self._access_keychain_item(service_name, account_name, timeout)
        except Exception as e:
            logging.debug(f"Touch ID authentication failed: {str(e)}")
            return None

    def _access_keychain_item(self, service_name: str, account_name: str, timeout_seconds: float) -> Optional[str]:
        """Access keychain item with authentication"""
        try:
            result = subprocess.run([
                'security', 'find-internet-password',
                '-s', service_name,
                '-a', account_name,
                '-w'
            ], capture_output=True, text=True, timeout=timeout_seconds)
            
            return result.stdout.strip() if result.returncode == 0 else None
        except Exception as e:
            logging.debug(f"Keychain access failed: {str(e)}")
            return None

    def store_credential(self, credential_id: str, private_key_data: bytes, rp_id: str, timeout_seconds: Optional[float] = None) -> bool:
        """Store private key in macOS keychain with Touch ID access control"""
        try:
            timeout = timeout_seconds or DEFAULT_BIOMETRIC_TIMEOUT
            encoded_key = base64.b64encode(private_key_data).decode('ascii')
            service_name = f"{self.service_prefix} - {rp_id}"
            account_name = f"webauthn-{credential_id}"
            
            success = self._store_with_touchid_access(service_name, account_name, encoded_key, rp_id, timeout)
            
            if success:
                self._set_touchid_access_control(service_name, account_name, timeout)
            
            return success
            
        except Exception as e:
            BiometricErrorHandler.create_storage_error("store", "macOS keychain", e)
            return False
    
    def _store_with_touchid_access(self, service_name: str, account_name: str, 
                                  encoded_key: str, rp_id: str, timeout_seconds: float) -> bool:
        """Store credential with Touch ID access control"""
        try:
            result = subprocess.run([
                'security', 'add-internet-password',
                '-s', service_name,
                '-a', account_name,
                '-w', encoded_key,
                '-D', 'WebAuthn Credential',
                '-j', f'Keeper biometric credential for {rp_id}',
                '-A',  # Allow access from any application
                '-T', '',  # No specific application restrictions
                '-U'  # Update if exists
            ], capture_output=True, text=True, timeout=timeout_seconds)
            
            if result.returncode == 0:
                return True
            
            return False
            
        except Exception as e:
            logging.warning(f"Failed to store credential: {str(e)}")
            return False
    
    def _set_touchid_access_control(self, service_name: str, account_name: str, timeout_seconds: float):
        """Set Touch ID access control for stored credential"""
        try:
            subprocess.run([
                'security', 'set-internet-password-partition-list',
                '-s', service_name,
                '-a', account_name,
                '-S', 'SmartCard,TouchID',
                '-k', ''
            ], capture_output=True, text=True, timeout=timeout_seconds)
            
        except Exception as e:
            logging.debug(f"Could not set Touch ID access control: {str(e)}")

    def load_credential(self, credential_id: str, rp_id: Optional[str] = None, timeout_seconds: Optional[float] = None) -> Optional[object]:
        """Load private key from macOS keychain using Touch ID"""
        try:
            timeout = timeout_seconds or DEFAULT_BIOMETRIC_TIMEOUT
            account_name = f"webauthn-{credential_id}"
            if not rp_id:
                raise Exception("RP ID is required for credential loading")
            service_names = [f"{self.service_prefix} - {rp_id}"]
            
            for service_name in service_names:
                encoded_key = self._load_from_service(service_name, account_name, timeout)
                if encoded_key:
                    key_data = base64.b64decode(encoded_key)
                    return crypto.load_ec_private_key(key_data)
            
            return None
            
        except Exception as e:
            BiometricErrorHandler.create_storage_error("load", "macOS keychain", e)
            return None
    
    def _load_from_service(self, service_name: str, account_name: str, timeout_seconds: float) -> Optional[str]:
        """Load credential from specific service"""
        try:
            result = subprocess.run([
                'security', 'find-internet-password',
                '-s', service_name,
                '-a', account_name,
                '-w'
            ], capture_output=True, text=True, timeout=timeout_seconds)
            
            if result.returncode == 0:
                return result.stdout.strip()
            
            if result.returncode == 44:
                return self._authenticate_with_touchid(service_name, account_name, timeout_seconds)
                
        except Exception:
            pass
        
        return None

    def delete_credential(self, credential_id: str, rp_id: Optional[str] = None, timeout_seconds: Optional[float] = None) -> bool:
        """Delete private key from macOS keychain"""
        try:
            timeout = timeout_seconds or DEFAULT_BIOMETRIC_TIMEOUT
            account_name = f"webauthn-{credential_id}"
            if not rp_id:
                raise Exception("RP ID is required for credential deletion")
            service_names = [f"{self.service_prefix} - {rp_id}"]
            
            success = False
            for service_name in service_names:
                try:
                    result = subprocess.run([
                        'security', 'delete-internet-password',
                        '-s', service_name,
                        '-a', account_name
                    ], capture_output=True, text=True, timeout=timeout)
                    
                    if result.returncode == 0:
                        success = True
                        
                except Exception:
                    continue
            
            return success
                    
        except Exception as e:
            BiometricErrorHandler.create_storage_error("delete", "macOS keychain", e)
            return False

    def credential_exists(self, credential_id: str, rp_id: Optional[str] = None, timeout_seconds: Optional[float] = None) -> bool:
        """Check if credential exists in macOS keychain"""
        try:
            timeout = timeout_seconds or DEFAULT_BIOMETRIC_TIMEOUT
            account_name = f"webauthn-{credential_id}"
            if not rp_id:
                raise Exception("RP ID is required for credential existence check")
            service_names = [f"{self.service_prefix} - {rp_id}"]
            
            for service_name in service_names:
                try:
                    result = subprocess.run([
                        'security', 'find-internet-password',
                        '-s', service_name,
                        '-a', account_name,
                        '-w'
                    ], capture_output=True, text=True, timeout=timeout)
                    
                    if result.returncode == 0:
                        return True
                        
                except Exception:
                    continue
            
            return False
            
        except Exception:
            return False 


#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2018 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import argparse
import json
import logging
import os
import re

from ..breachwatch import BreachWatch

from .base import suppress_exit, raise_parse_exception, Command
from .. import api, generator, utils, crypto
from ..error import CommandError
from ..params import LAST_RECORD_UID
from ..record import Record
from ..subfolder import BaseFolderNode, try_resolve_path


add_parser = argparse.ArgumentParser(prog='add', description='Add a record')
add_parser.add_argument('--login', dest='login', action='store', help='login name')
add_parser.add_argument('--pass', dest='password', action='store', help='password')
add_parser.add_argument('--url', dest='url', action='store', help='url')
add_parser.add_argument('--notes', dest='notes', action='store', help='notes')
add_parser.add_argument('--custom', dest='custom', action='store', help='add custom fields. JSON or name:value pairs separated by comma. CSV Example: --custom "name1: value1, name2: value2". JSON Example: --custom \'{"name1":"value1", "name2":"value: 2,3,4"}\'')
add_parser.add_argument('--folder', dest='folder', action='store', help='folder path or UID where record is to be created')
add_parser.add_argument('-f', '--force', dest='force', action='store_true', help='do not prompt for omitted fields')
add_parser.add_argument('-g', '--generate', dest='generate', action='store_true', help='generate a random password')
add_parser.add_argument('title', type=str, action='store', help='record title')
add_parser.error = raise_parse_exception
add_parser.exit = suppress_exit


edit_parser = argparse.ArgumentParser(prog='edit', description='Edit a record')
edit_parser.add_argument('--login', dest='login', action='store', help='login name')
edit_parser.add_argument('--pass', dest='password', action='store', help='password')
edit_parser.add_argument('--url', dest='url', action='store', help='url')
edit_parser.add_argument('--notes', dest='notes', action='store', help='set or replace the notes. Use a plus sign (+) in front appends to existing notes')
edit_parser.add_argument('--custom', dest='custom', action='store', help='custom fields. JSON or name:value pairs separated by comma. CSV Example: --custom "name1: value1, name2: value2". JSON Example: --custom \'{"name1":"value1", "name2":"value: 2,3,4"}\'')
edit_parser.add_argument('-g', '--generate', dest='generate', action='store_true', help='generate a random password')
edit_parser.add_argument('record', nargs='?', type=str, action='store', help='record path or UID')
edit_parser.error = raise_parse_exception
edit_parser.exit = suppress_exit


class RecordUtils(object):
    parameter_pattern = re.compile(r'^\${([^:]+?):([^}]+?)}$')

    @staticmethod
    def custom_field_value(value):  # type: (any) -> str
        if not value:
            return ''
        if type(value) != str:
            return value
        m = RecordUtils.parameter_pattern.match(value.strip())
        if m:
            parts = m.groups()
            if len(parts) == 2:
                if parts[0].lower() == 'file':
                    filename = parts[1].strip()
                    if os.path.isfile(filename):
                        with open(filename, 'r') as f:
                            return f.read()
                elif parts[0].lower() == 'env':
                    if parts[1] in os.environ:
                        return os.environ[parts[1]]

        return value


class RecordAddCommand(Command, RecordUtils):
    def get_parser(self):
        return add_parser

    def execute(self, params, **kwargs):
        title = kwargs['title'] if 'title' in kwargs else None
        login = kwargs['login'] if 'login' in kwargs else None
        password = kwargs['password'] if 'password' in kwargs else None
        url = kwargs['url'] if 'url' in kwargs else None
        custom_list = kwargs['custom'] if 'custom' in kwargs else None
        notes = kwargs['notes'] if 'notes' in kwargs else None

        generate = kwargs['generate'] if 'generate' in kwargs else None
        if generate:
            password = generator.generate(16)

        force = kwargs['force'] if 'force' in kwargs else None
        if not force:
            if not title:
                title = input('...' + 'Title: '.rjust(16))
            if not login:
                login = input('...' + 'Login: '.rjust(16))
            if not password:
                password = input('...' + 'Password: '.rjust(16))
            if not url:
                url = input('...' + 'Login URL: '.rjust(16))
        if not title:
            raise CommandError('add', 'Invalid title. Expected non-empty string.')

        custom = []
        if custom_list:
            if type(custom_list) == str:
                if custom_list[0] == '{' and custom_list[-1] == '}':
                    try:
                        custom_json = json.loads(custom_list)
                        for k, v in custom_json.items():
                            custom.append({
                                'name': k,
                                'value': self.custom_field_value(v)
                            })
                    except ValueError as e:
                        raise CommandError('add', 'Invalid custom fields JSON input: {0}'.format(e))
                else:
                    pairs = custom_list.split(',')
                    for pair in pairs:
                        idx = pair.find(':')
                        if idx > 0:
                            custom.append({
                                'name': pair[:idx].strip(),
                                'value': self.custom_field_value(pair[idx+1:].strip())
                            })
                        else:
                            raise CommandError('add', 'Invalid custom fields input. Expected: "Key:Value". Got: "{0}"'.format(pair))

            elif type(custom_list) == list:
                for c in custom_list:
                    if type(c) == dict:
                        name = c.get('name')
                        value = c.get('value')
                        if name and value:
                            custom.append({
                                'type': 'text',
                                'name': name,
                                'value': self.custom_field_value(value)
                            })

        folder = None
        folder_name = kwargs['folder'] if 'folder' in kwargs else None
        if folder_name:
            if folder_name in params.folder_cache:
                folder = params.folder_cache[folder_name]
            else:
                src = try_resolve_path(params, folder_name)
                if src is not None:
                    folder, name = src
                    if name:
                        raise CommandError('add', 'No such folder: {0}'.format(folder_name))
                else:
                    raise CommandError('add', 'No such folder: {0}'.format(folder_name))
        if not folder:
            folder = params.folder_cache[params.current_folder] if params.current_folder else params.root_folder

        if not force:
            folder_uid = folder.uid or ''
            if folder_uid in params.subfolder_record_cache:
                for uid in params.subfolder_record_cache[folder_uid]:
                    r = api.get_record(params, uid)
                    if r.title == title:
                        raise CommandError('add', 'Record with title "{0}" already exists'.format(title))

        record_key = os.urandom(32)
        record_uid = api.generate_record_uid()
        rq = {
            'command': 'record_add',
            'record_uid': record_uid,
            'record_type': 'password',
            'record_key': utils.base64_url_encode(crypto.encrypt_aes_v1(record_key, params.data_key)),
            'how_long_ago': 0
        }
        if folder.type in {BaseFolderNode.SharedFolderType, BaseFolderNode.SharedFolderFolderType}:
            rq['folder_uid'] = folder.uid
            rq['folder_type'] = 'shared_folder' if folder.type == BaseFolderNode.SharedFolderType else 'shared_folder_folder'

            sh_uid = folder.uid if folder.type == BaseFolderNode.SharedFolderType else folder.shared_folder_uid
            sf = params.shared_folder_cache[sh_uid]
            rq['folder_key'] = utils.base64_url_encode(
                crypto.encrypt_aes_v1(record_key, sf['shared_folder_key_unencrypted']))
            if 'key_type' not in sf:
                if 'teams' in sf:
                    for team in sf['teams']:
                        rq['team_uid'] = team['team_uid']
                        if team['manage_records']:
                            break
        else:
            rq['folder_type'] = 'user_folder'
            if folder.type != BaseFolderNode.RootFolderType:
                rq['folder_uid'] = folder.uid

        data = {
            'title': title or '',
            'secret1': login or '',
            'secret2': password or '',
            'link': url or '',
            'notes': notes or '',
            'custom': custom or []
        }
        Record.validate_record_data(data, None, None)

        rq['data'] = utils.base64_url_encode(crypto.encrypt_aes_v1(json.dumps(data).encode('utf-8'), record_key))

        api.communicate(params, rq)
        api.sync_down(params)
        if params.enterprise_ec_key:
            api.add_record_audit_data(params, [record_uid])
        BreachWatch.scan_and_update_security_data(params, record_uid, params.breach_watch)

        params.sync_data = True
        params.environment_variables[LAST_RECORD_UID] = record_uid
        return record_uid


class RecordEditCommand(Command, RecordUtils):
    def get_parser(self):
        return edit_parser

    def execute(self, params, **kwargs):
        name = kwargs.get('record')

        if not name:
            self.get_parser().print_help()
            return

        record_uid = None
        if name in params.record_cache:
            record_uid = name
        else:
            rs = try_resolve_path(params, name)
            if rs is not None:
                folder, name = rs
                if folder is not None and name is not None:
                    folder_uid = folder.uid or ''
                    if folder_uid in params.subfolder_record_cache:
                        for uid in params.subfolder_record_cache[folder_uid]:
                            r = api.get_record(params, uid)
                            if r.title.lower() == name.lower():
                                record_uid = uid
                                break

        if record_uid is None:
            raise CommandError('edit', 'Enter name or uid of existing record')

        record = api.get_record(params, record_uid)

        changed = False
        password_changed = False
        if kwargs.get('title') is not None:
            title = kwargs['title']
            if title:
                record.title = title
                changed = True
            else:
                logging.warning('Record title cannot be empty.')
        if kwargs.get('login') is not None:
            record.login = kwargs['login']
            changed = True
        if kwargs.get('password') is not None:
            last_password = record.unmasked_password or record.password
            record.password = kwargs['password']
            password_changed = record.password != last_password
            changed = True
        else:
            if kwargs.get('generate'):
                record.password = generator.generate(16)
                changed = password_changed = True
        if kwargs.get('url') is not None:
            record.login_url = kwargs['url']
            changed = True
        if kwargs.get('notes') is not None:
            notes = kwargs['notes'] # type: str
            if notes:
                if notes.startswith("+"):
                    notes = record.notes + "\n" + notes[1:]
            record.notes = notes
            changed = True

        custom_list = kwargs.get('custom')
        if custom_list:
            custom = []
            if type(custom_list) == str:
                if custom_list[0] == '{' and custom_list[-1] == '}':
                    try:
                        custom_json = json.loads(custom_list)
                        for k,v in custom_json.items():
                            custom.append({
                                'type': 'text',
                                'name': k,
                                'value': self.custom_field_value(v)
                            })
                    except ValueError as e:
                        raise CommandError('edit', 'Invalid custom fields JSON input: {0}'.format(e))
                else:
                    pairs = custom_list.split(',')
                    for pair in pairs:
                        idx = pair.find(':')
                        if idx > 0:
                            custom.append({
                                'name': pair[:idx].strip(),
                                'value': self.custom_field_value(pair[idx+1:].strip())
                            })
                        else:
                            raise CommandError('edit', 'Invalid custom fields input. Expected: "Key:Value". Got: "{0}"'.format(pair))
            elif type(custom_list) == list:
                for c in custom_list:
                    if type(c) == dict:
                        name = c.get('name')
                        value = c.get('value')
                        if name and value:
                            custom.append({
                                'name': name,
                                'value': self.custom_field_value(value)
                            })
            if custom:
                for c in custom:
                    if c['value']:
                        record.set_field(c['name'], c['value'])
                        changed = True
                    else:
                        deleted = record.remove_field(c['name'])
                        if deleted:
                            changed = True

        if changed:
            api.update_record(params, record)
            if password_changed:
                BreachWatch.scan_and_update_security_data(params, record_uid, params.breach_watch)
            params.sync_data = True



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2025 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import json
import logging
import getpass
import threading
from datetime import datetime, timedelta
from typing import Tuple

from . import api, utils, crypto
from .proto import APIRequest_pb2
from .display import bcolors
from .params import KeeperParams
from .error import KeeperApiError


class MasterPasswordReentryEnforcer:
    """Handler for MASTER_PASSWORD_REENTRY enforcement with multi-factor authentication support."""
    
    # Class variable to store the last successful validation time per user (thread-safe)
    _last_validation_time = {}
    _validation_lock = threading.RLock()
    
    @classmethod
    def requires_master_password_reentry(cls, params: KeeperParams, operation: str = "record_level") -> bool:
        """
        Check if master password reentry is required for the given operation.
        
        Args:
            params: KeeperParams instance
            operation: The operation being performed (default: "record_level")
            
        Returns:
            bool: True if master password reentry is required
        """
        # Input validation for operation parameter
        if not operation or not isinstance(operation, str) or len(operation.strip()) == 0:
            logging.error("Invalid operation parameter provided")
            return False
            
        # Sanitize operation string to prevent injection attacks
        operation = operation.strip()[:100]  # Limit length and strip whitespace
        # Bypass enforcement when running in service mode
        if params and hasattr(params, 'service_mode') and params.service_mode:
            logging.info(f"Bypassing master password enforcement for operation '{operation}' - running in service mode")
            return False

        if not params or not params.enforcements:
            return False
            
        # Look for MASTER_PASSWORD_REENTRY enforcement
        json_enforcements = params.enforcements.get('jsons', [])
        
        for enforcement in json_enforcements:
            if enforcement.get('key') == 'master_password_reentry':
                try:
                    enforcement_value = json.loads(enforcement.get('value', '{}'))
                    operations = enforcement_value.get('operations', [])
                    timeout_minutes = enforcement_value.get('timeout', 5)
                    
                    # Validate timeout to prevent integer overflow and ensure reasonable bounds
                    if not isinstance(timeout_minutes, (int, float)) or timeout_minutes < 0 or timeout_minutes > 999:  # Max 999 minutes
                        logging.error("Invalid timeout value in enforcement policy, timeout value must be more than or equal to 0 and less than or equal to 999 minutes, timeout value will be set to 5 minutes")
                        timeout_minutes = 5  # Default fallback
                    
                    # Check if the current operation requires master password reentry
                    if operation in operations:
                        # Check if we're still within the timeout period (thread-safe)
                        user_key = params.user if params.user else None
                        if not user_key:
                            logging.error("User key not available")
                            return True  # Enforcement required if user key is missing
                            
                        with cls._validation_lock:
                            if user_key in cls._last_validation_time:
                                last_validation = cls._last_validation_time[user_key]
                                if datetime.now() - last_validation < timedelta(minutes=timeout_minutes):
                                    return False  # Still within timeout, no need to reenter
                        
                        return True  # Enforcement required
                        
                except (json.JSONDecodeError, KeyError, TypeError):
                    logging.error("Failed to parse MASTER_PASSWORD_REENTRY enforcement")
                    
        return False

    @classmethod
    def _is_biometric_available(cls, params: KeeperParams) -> bool:
        """
        Check if biometric authentication is available for the user.
        
        Args:
            params: KeeperParams instance
            
        Returns:
            bool: True if biometric authentication is available
        """
        try:
            from .biometric import check_biometric_previously_used, BiometricClient
            from .biometric.platforms.detector import BiometricDetector
            
            # Check if biometric was previously used
            if not check_biometric_previously_used(params.user):
                return False
                
            # Check if platform supports biometric
            detector = BiometricDetector()
            supported, _ = detector.detect_platform_capabilities()
            if not supported:
                return False
                
            # Check if user has registered biometric credentials
            client = BiometricClient()
            credentials = client.get_available_credentials(params)
            return len(credentials) > 0
            
        except (ImportError, Exception):
            logging.debug("Biometric not available")
            return False
    
    @classmethod 
    def _validate_biometric(cls, params: KeeperParams) -> bool:
        """
        Perform biometric authentication for vault re-authentication.
        
        Args:
            params: KeeperParams instance
            
        Returns:
            bool: True if biometric validation successful, False otherwise
        """
        try:
            from .biometric import BiometricClient
            
            print(f'{bcolors.WARNING}Biometric authentication required for this operation.{bcolors.ENDC}')
            print('Please complete biometric authentication...')
            
            client = BiometricClient()
            
            # Generate authentication options for vault re-authentication
            auth_options = client.generate_authentication_options(params, purpose='vault')
            
            # Perform biometric authentication
            auth_response = client.perform_authentication(auth_options)
            
            if auth_response:
                # Store successful validation time (thread-safe)
                user_key = params.user if params and params.user else None
                if user_key:
                    with cls._validation_lock:
                        cls._last_validation_time[user_key] = datetime.now()
                print(f'{bcolors.OKGREEN}Biometric authentication successful.{bcolors.ENDC}')
                return True
            else:
                logging.warning("Biometric authentication failed")
                return False
                
        except KeyboardInterrupt:
            logging.info("Biometric authentication cancelled by user")
            return False
        except Exception:
            logging.debug("Biometric authentication failed")
            return False
    
    @classmethod
    def _has_master_password_or_alternate(cls, params: KeeperParams) -> Tuple[bool, bool]:
        """
        Check if user has master password or alternate SSO master password.
        
        Args:
            params: KeeperParams instance
            
        Returns:
            Tuple[bool, bool]: (has_regular_master_password, has_alternate_sso_password)
        """
        is_sso_user = params.settings.get('sso_user', False) if params.settings else False
        logging.debug(f"Checking authentication methods - is_sso_user: {is_sso_user}")
        
        # For regular users, check if they can use master password authentication
        has_salt = hasattr(params, 'salt') and params.salt
        has_iterations = hasattr(params, 'iterations') and params.iterations
        
        # If user is logged in (has session token) but doesn't have salt/iterations in params,
        # assume they can still do master password auth (persistent login scenario)
        # We'll fetch salt/iterations when needed during validation
        if not is_sso_user:
            if has_salt and has_iterations:
                has_regular = True
                logging.debug(f"Regular user - salt/iterations available in params")
            elif params.session_token and params.user:
                has_regular = True
                logging.debug(f"Regular user - persistent login detected, will fetch salt/iterations when needed")
            else:
                has_regular = False
                logging.debug(f"Regular user - no session or authentication data available")
        else:
            has_regular = False
        
        # Check if SSO user has alternate password
        has_alternate = False
        if is_sso_user:
            try:
                current_salt = api.communicate_rest(params, None, 'authentication/get_salt_and_iterations',
                                                  rs_type=APIRequest_pb2.Salt)
                has_alternate = current_salt is not None
                logging.debug(f"SSO user - API call successful")
            except KeeperApiError as kae:
                has_alternate = kae.result_code != 'doesnt_exist'
                logging.debug(f"SSO user - API error: {kae.result_code}")
            except Exception:
                logging.debug("SSO user - Exception checking alternate password")
                pass
                
        return has_regular, has_alternate
    
    @classmethod
    def _validate_master_or_alternate_password(cls, params: KeeperParams) -> bool:
        """
        Validate master password or alternate SSO master password.
        
        Args:
            params: KeeperParams instance
            
        Returns:
            bool: True if validation successful, False otherwise
        """
        is_sso_user = params.settings.get('sso_user', False)
        has_regular, has_alternate = cls._has_master_password_or_alternate(params)
        
        if not has_regular and not has_alternate:
            return False
            
        try:
            if is_sso_user and has_alternate:
                # Get salt and iterations for alternate password
                try:
                    current_salt = api.communicate_rest(params, None, 'authentication/get_salt_and_iterations',
                                                      rs_type=APIRequest_pb2.Salt)
                    salt = current_salt.salt
                    iterations = current_salt.iterations
                except Exception:
                    logging.error("Cannot validate alternate SSO master password: no salt information available")
                    return False
                    
                prompt_text = f'{bcolors.WARNING}Alternate SSO master password reentry required for this operation.{bcolors.ENDC}\nEnter alternate SSO master password: '
            else:
                # Use regular master password
                # Check if salt/iterations are in params, if not fetch them (persistent login scenario)
                if hasattr(params, 'salt') and params.salt and hasattr(params, 'iterations') and params.iterations:
                    salt = params.salt
                    iterations = params.iterations
                    logging.debug("Using salt/iterations from params")
                else:
                    # Fetch salt/iterations from server (persistent login case)
                    try:
                        logging.debug("Fetching salt/iterations from server for persistent login user")
                        current_salt = api.communicate_rest(params, None, 'authentication/get_salt_and_iterations',
                                                          rs_type=APIRequest_pb2.Salt)
                        salt = current_salt.salt
                        iterations = current_salt.iterations
                        logging.debug("Successfully fetched salt/iterations from server")
                    except Exception:
                        logging.error("Cannot validate master password: failed to get salt information from server")
                        return False
                    
                prompt_text = f'{bcolors.WARNING}Master password reentry required for this operation.{bcolors.ENDC}\nEnter master password: '
            
            # Prompt for password
            master_password = getpass.getpass(prompt=prompt_text).strip()
            
            if not master_password:
                logging.error("Password is required")
                return False
            
            try:
                # Create authentication hash
                auth_hash = crypto.derive_keyhash_v1(master_password, salt, iterations)
                
                # Create master password reentry request
                rq = APIRequest_pb2.MasterPasswordReentryRequest()
                rq.pbkdf2Password = utils.base64_url_encode(auth_hash)
                rq.action = APIRequest_pb2.UNMASK
                
                # Validate with server
                rs = api.communicate_rest(params, rq, 'authentication/validate_master_password',
                                        rs_type=APIRequest_pb2.MasterPasswordReentryResponse, payload_version=1)
                
                if rs.status == APIRequest_pb2.MP_SUCCESS:
                    # Store successful validation time (thread-safe)
                    user_key = params.user if params and params.user else None
                    if user_key:
                        with cls._validation_lock:
                            cls._last_validation_time[user_key] = datetime.now()
                    return True
                else:
                    logging.error("Password validation failed")
                    return False
                    
            finally:
                # Clear sensitive data from memory
                master_password = 'x' * len(master_password)
                del master_password
                if 'auth_hash' in locals():
                    auth_hash = b'x' * len(auth_hash)
                    del auth_hash
                
        except Exception:
            logging.error("Password validation failed")
            return False
    
    @classmethod
    def check_and_enforce(cls, params: KeeperParams, operation: str = "record_level") -> bool:
        """
        Check if enforcement is required and handle multi-factor authentication if needed.
        
        Uses priority-based authentication:
        1. First check biometric (if enabled)
        2. Then check master password/alternate SSO master password (if available)
        3. If none available, policy does not apply
        
        Args:
            params: KeeperParams instance
            operation: The operation being performed
            
        Returns:
            bool: True if operation can proceed, False if validation failed
        """
        if not cls.requires_master_password_reentry(params, operation):
            return True
            
        # Priority 1: Biometric authentication
        if cls._is_biometric_available(params):
            logging.info("Using biometric authentication for re-authentication")
            return cls._validate_biometric(params)
        
        # Priority 2: Master password or alternate SSO master password
        has_regular, has_alternate = cls._has_master_password_or_alternate(params)
        if has_regular or has_alternate:
            if has_alternate:
                logging.info("Using alternate SSO master password for re-authentication")
            else:
                logging.info("Using master password for re-authentication")
            return cls._validate_master_or_alternate_password(params)
        
        # If no authentication methods are available, policy does not apply
        logging.info("Master Password Re-authentication skipped: Master Password or alternate SSO Master Password not available - policy does not apply")
        return True



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2022 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import argparse
import logging
import re
from typing import Dict, List

from tabulate import tabulate

from . import plugin_manager
from .. import api, record_management
from .. import generator
from ..commands.base import raise_parse_exception, suppress_exit, Command
from ..params import KeeperParams
from ..subfolder import find_folders, get_folder_path
from ..utils import confirm
from ..vault import KeeperRecord, CustomField, TypedField, PasswordRecord, TypedRecord


def register_commands(commands):
    commands['rotate'] = RecordRotateCommand()


def register_command_info(aliases, command_info):
    aliases['r'] = 'rotate'
    for p in [rotate_parser]:
        command_info[p.prog] = p.description


rotate_parser = argparse.ArgumentParser(
    prog='rotate', description='Rotate the password for a Keeper record from this Commander.'
)
rotate_parser.add_argument(
    '--print', dest='print', action='store_true', help='display the record content after rotation'
)
rotate_parser.add_argument(
    '-m', '--match', dest='match', action='store', help='regular expression to select records for password rotation'
)
rotate_parser.add_argument('--plugin', dest='plugin', action='store', help='Force rotation plugin (optional)')
rotate_parser.add_argument('--host', dest='host', action='store', help='Optional host (override record value)')
rotate_parser.add_argument('--port', dest='port', action='store', help='Optional port (override record value)')
rotate_parser.add_argument('-r', '--rules', dest='rules', action='store', help='Optional rules (override record value)')
rotate_parser.add_argument(
    '-l', '--length', type=int, dest='length', action='store', help='Optional password length (override record value)'
)
rotate_parser.add_argument('--password', dest='password', action='store', help='Optional new password')
rotate_parser.add_argument(
    '--force', dest='force', action='store_true', help='force all matches to rotate without prompt'
)
rotate_parser.add_argument(
    'name', nargs='?', type=str, action='store', help='record UID or name assigned to rotate command'
)
rotate_parser.error = raise_parse_exception
rotate_parser.exit = suppress_exit


def adjust_password(password):   # type: (str) -> str
    if not password:
        return password
    if password[0].isalnum():
        return password
    for i in range(1, len(password)):
        if password[i].isalnum():
            return password[i] + password[1:i] + password[0] + password[i+1:]
    return 'a' + password


def update_custom_text_fields(record, new_field_dict):
    fld_attr = plugin_manager.get_custom_field_attr(record)
    for f in record.custom:
        field_name = getattr(f, fld_attr)
        if field_name in new_field_dict:
            if isinstance(record, TypedRecord):
                f.value[0] = new_field_dict.pop(field_name)
            elif isinstance(record, PasswordRecord):
                f.value = new_field_dict.pop(field_name)
    for k, v in new_field_dict.items():
        if isinstance(record, TypedRecord):
            record.custom.append(TypedField({'type': 'text', 'label': k, 'value': [v]}))
        elif isinstance(record, PasswordRecord):
            record.custom.append(CustomField({'type': 'text', 'name': k, 'value': v}))


def update_password(params, record, new_password, plugin):
    if hasattr(plugin, 'update_password'):
        plugin.update_password(record, new_password)
    elif isinstance(record, PasswordRecord):
        record.password = new_password
    elif isinstance(record, TypedRecord):
        password_field = record.get_typed_field('password')
        if password_field is None:
            logging.warning('Creating new password for record because existing password is not found.')
            record.custom.append(TypedField({'type': 'password', 'value': [new_password]}))
        else:
            password_field.value = [new_password]
    else:
        logging.error(f'Record {record.title} is an invalid type of record for rotation update')
        return False

    try:
        record_management.update_record(params, record)
    except Exception as e:
        logging.error(e)
        return False
    else:
        return True


def get_new_password(plugin, rules=None, length=None):
    if hasattr(plugin, 'disallow_special_characters'):
        pw_special_characters = generator.PW_SPECIAL_CHARACTERS.translate(
            str.maketrans('', '', plugin.disallow_special_characters)
        )
    else:
        pw_special_characters = generator.PW_SPECIAL_CHARACTERS
    if rules:
        logging.debug("Rules found for record")
        kpg = generator.KeeperPasswordGenerator.create_from_rules(
            rules, length=length, special_characters=pw_special_characters
        )
        if kpg is None:
            logging.warning('Using default password complexity rules')
            if length is None:
                length = generator.DEFAULT_PASSWORD_LENGTH
            kpg = generator.KeeperPasswordGenerator(length=length, special_characters=pw_special_characters)
    else:
        logging.debug("No rules, just generate")
        if length is None:
            length = generator.DEFAULT_PASSWORD_LENGTH
        kpg = generator.KeeperPasswordGenerator(length=length, special_characters=pw_special_characters)
    new_password = kpg.generate()

    # ensure password starts with alpha numeric character
    new_password = adjust_password(new_password)

    # Some plugins might need to change the password in the process of rotation
    # f.e. windows plugin gets rid of certain characters.
    if hasattr(plugin, "adjust"):
        new_password = plugin.adjust(new_password)

    return new_password


def rotate_password(params, record_uid, rotate_name=None, plugin_name=None, host=None, port=None, rules=None,
                    length=None, new_password=None):
    """ Rotate the password for the specified record """
    api.sync_down(params)
    record = KeeperRecord.load(params, record_uid)
    if api.resolve_record_write_path(params, record_uid) is None:
        logging.error(
            f'The rotation failed for record "{record.title}" (uid=[{record.record_uid}]):\n'
            'The target record is not editable but needs to be updated with new credentials to complete the rotation.'
        )
        return False

    rs = plugin_manager.get_plugin(record, rotate_name, plugin_name, host, port)
    if not isinstance(rs, tuple):
        return False
    if len(rs) != 3:
        return False

    plugin_name, plugin_kwargs, plugin = rs
    if not plugin:
        return False

    if new_password is None:
        if not rules:
            rules = plugin_kwargs.get('rules')
        if not length:
            length = plugin_kwargs.get('length')
        new_password = get_new_password(plugin, rules, length)

    if plugin_kwargs.get('password') == new_password:
        logging.warning('Rotation aborted because the old and new passwords are the same.')
        success = False
    else:
        if hasattr(plugin, 'rotate_start_msg'):
            plugin.rotate_start_msg()
        else:
            logging.info(f'Rotating with plugin {plugin_name}')
        try:
            success = plugin.rotate(record, new_password)
        except Exception as e:
            logging.info('Error rotating with plugin %s: %s', plugin_name, e)
            success = False

    if success:
        logging.debug(f'Password rotation successful for "{plugin_name}".')
    else:
        logging.warning(
            f'Password rotation failed for record "{record.title}" (uid=[{record.record_uid}]), plugin "{plugin_name}".'
        )
        return False

    if update_password(params, record, new_password, plugin):
        api.sync_down(params)
        new_record = KeeperRecord.load(params, record_uid)
        if hasattr(plugin, 'sync_password'):
            plugin.sync_password()
        logging.info(f'Password rotation successful for record "{new_record.title}".')
        return True
    elif hasattr(plugin, 'revert') and plugin.revert(record, new_password):
        logging.warning(
            f'Unable to update the record "{record.title}" (uid=[{record.record_uid}]), so the rotation was reverted.'
        )
    else:
        if hasattr(plugin, 'revert_failed_msg'):
            plugin.revert_failed_msg()
        else:
            logging.error(
                f"Rotated to new password {new_password} but couldn't update the record "
                f'"{record.title}" (uid=[{record.record_uid}]). The new password will be needed for access.'
            )

    return False


class RotateEndpoint:
    def __init__(self, name, type, description, record_uid, record_title, paths):
        self.name = name
        self.type = type
        self.description = description
        self.record_uid = record_uid
        self.record_title = record_title
        self.paths = paths


rotate_pattern = re.compile(r'^cmdr:plugin(:[^:]*)?$')
rotate_desc_pattern = re.compile(r'^cmdr:plugin:([^:]+):description$')


class RecordRotateCommand(Command):
    def get_parser(self):
        return rotate_parser

    def execute(self, params, **kwargs):
        print_result = kwargs['print'] if 'print' in kwargs else None
        name = kwargs['name'] if 'name' in kwargs else None
        match = kwargs['match'] if 'match' in kwargs else None
        force = kwargs['force'] if 'force' in kwargs else None
        if name:
            record_uid = None
            rotate_name = None
            if name in params.record_cache:
                record_uid = name
            else:
                RecordRotateCommand.find_endpoints(params)
                nl = name.lower()
                endpoints = [
                    x for x in RecordRotateCommand.Endpoints if x.record_title.lower() == nl or x.name.lower() == nl
                ]
                if len(endpoints) > 0:
                    if len(endpoints) == 1:
                        record_uid = endpoints[0].record_uid
                        rotate_name = endpoints[0].name
                    else:
                        logging.error('There are more than one rotation records with name %s. Please use record UID.', name)
                        return
            if record_uid:
                rotate_password(
                    params, record_uid, rotate_name=rotate_name, plugin_name=kwargs.get('plugin'),
                    host=kwargs.get('host'), port=kwargs.get('port'), rules=kwargs.get('rules'),
                    length=kwargs.get('length'), new_password=kwargs.get('password')
                )
                if print_result:
                    record = api.get_record(params, record_uid)
                    record.display()
            else:
                logging.error('Rotate {0}: not found'.format(name))
        elif match:
            results = api.search_records(params, match)
            for r in results:
                if r.version not in (2, 3):
                    continue
                if force or confirm(f'Rotate password for record {r.title}?'):
                    rotate_password(
                        params, r.record_uid, plugin_name=kwargs.get('plugin'), new_password=kwargs.get('password'),
                        host=kwargs.get('host'), port=kwargs.get('port'), rules=kwargs.get('rules')
                    )
                    if print_result:
                        record = api.get_record(params, r.record_uid)
                        record.display()
        else:
            RecordRotateCommand.find_endpoints(params)
            if RecordRotateCommand.Endpoints:
                logging.info("Available records for password rotation")
                logging.info('')
                headers = ["#", 'Name', 'Type', 'Record UID', 'Record Title', 'Folder(s)']
                table = []
                for i in range(len(RecordRotateCommand.Endpoints)):
                    endpoint = RecordRotateCommand.Endpoints[i]
                    title = endpoint.record_title
                    folder = endpoint.paths[0] if len(endpoint.paths) > 0 else '/'
                    if len(title) > 23:
                        title = title[:20] + '...'
                    table.append([i + 1, endpoint.name, endpoint.type, endpoint.record_uid,  title, folder])
                print(tabulate(table, headers=headers))
                print('')

    LastRevision = 0
    Endpoints = []    # type: List[RotateEndpoint]

    @staticmethod
    def find_endpoints(params):       # type: (KeeperParams) -> None
        if RecordRotateCommand.LastRevision < params.revision:
            RecordRotateCommand.LastRevision = params.revision
            RecordRotateCommand.Endpoints.clear()
            for rec in params.record_cache.values():
                if rec.get('version', 0) not in {2, 3}:
                    continue
                record = KeeperRecord.load(params, rec)
                endpoints = {}    # type: Dict[str, str]
                endpoints_desc = {}

                if isinstance(record, PasswordRecord):
                    for cf in record.custom or []:
                        m = rotate_pattern.match(cf.name)
                        if m and cf.value:
                            endpoints[cf.name] = cf.value
                        else:
                            m = rotate_desc_pattern.match(cf.name)
                            if m:
                                endpoints_desc[m[1]] = cf.value
                elif isinstance(record, TypedRecord):
                    for cf in record.custom or []:
                        if not cf.label:
                            continue
                        value = cf.get_default_value(str)
                        if not value:
                            continue
                        m = rotate_pattern.match(cf.label)
                        if m:
                            endpoints[cf.label] = value
                        else:
                            m = rotate_desc_pattern.match(cf.label)
                            if m:
                                endpoints_desc[m[1]] = value
                if endpoints:
                    paths = []
                    for folder_uid in find_folders(params, record.record_uid):
                        path = '/' + get_folder_path(params, folder_uid, '/')
                        paths.append(path)
                    for endpoint in endpoints:
                        name = endpoint
                        if name.startswith('cmdr:plugin'):
                            name = name[len('cmdr:plugin'):]
                            if name and name[0] == ':':
                                name = name[1:]
                        epoint = RotateEndpoint(name, endpoints[endpoint], endpoints_desc.get(endpoint) or '',
                                                record.record_uid, record.title, paths)
                        RecordRotateCommand.Endpoints.append(epoint)
            RecordRotateCommand.Endpoints.sort(key=lambda x: x.name)



import sys
if sys.version_info >= (3, 8):
    from unittest import TestCase, mock
    from keepercommander.params import KeeperParams
    from keepercommander.service.config.service_config import ServiceConfig
    from keepercommander.service.commands.config_operation import AddConfigService

    class TestConfigOperation(TestCase):
        def setUp(self):
            self.mock_params = mock.Mock(spec=KeeperParams)
            self.cmd = AddConfigService()
            
        def test_execute_with_existing_config(self):
            mock_config = {
                "is_advanced_security_enabled": "y",
                "records": []
            }
            mock_record = {
                "api-key": "test-api-key",
                "command_list": "list",
                "expiration_timestamp": "2024-12-31T23:59:59",
                #"expiration_of_token": ""
            }
            
            with mock.patch.object(ServiceConfig, 'load_config', return_value=mock_config), \
                mock.patch.object(ServiceConfig, 'create_record', return_value=mock_record), \
                mock.patch.object(ServiceConfig, 'save_config') as mock_save, \
                mock.patch('builtins.print'):
                
                self.cmd.execute(self.mock_params)
                
                expected_config = {
                    "is_advanced_security_enabled": "y",
                    "records": [mock_record]
                }
                mock_save.assert_called_once_with(expected_config)

        def test_execute_when_config_not_found(self):
            with mock.patch.object(ServiceConfig, 'load_config', side_effect=FileNotFoundError), \
                mock.patch('builtins.print') as mock_print:
                
                result = self.cmd.execute(self.mock_params)
                
                mock_print.assert_called_with(
                    "Error: Service configuration file not found. Please use 'service-create' command to create a service_config file."
                )
                self.assertEqual(result, '')

        def test_execute_with_general_error(self):
            with mock.patch.object(ServiceConfig, 'load_config', side_effect=Exception("Test error")), \
                mock.patch('builtins.print') as mock_print:
                
                result = self.cmd.execute(self.mock_params)
                
                mock_print.assert_called_with(
                    "Error: Service configuration file not found. Please use 'service-create' command to create a service_config file."
                )
                self.assertEqual(result, '')

        def test_create_and_add_record(self):
            mock_config = {
                "is_advanced_security_enabled": "n",
                "records": [{"existing": "record"}]
            }
            mock_record = {
                "api-key": "new-api-key",
                "command_list": "list",
                "expiration_timestamp": "2024-12-31T23:59:59",
                #"expiration_of_token": ""
            }
            
            with mock.patch.object(ServiceConfig, 'load_config', return_value=mock_config), \
                mock.patch.object(ServiceConfig, 'create_record', return_value=mock_record), \
                mock.patch.object(ServiceConfig, 'save_config') as mock_save, \
                mock.patch('builtins.print'):
                
                self.cmd.execute(self.mock_params)
                
                expected_config = {
                    "is_advanced_security_enabled": "n",
                    "records": [{"existing": "record"}, mock_record]
                }
                mock_save.assert_called_once_with(expected_config)


from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class BreachWatchInfoType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []
    RECORD: _ClassVar[BreachWatchInfoType]
    ALTERNATE_PASSWORD: _ClassVar[BreachWatchInfoType]

class BWStatus(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []
    GOOD: _ClassVar[BWStatus]
    CHANGED: _ClassVar[BWStatus]
    WEAK: _ClassVar[BWStatus]
    BREACHED: _ClassVar[BWStatus]
    IGNORE: _ClassVar[BWStatus]
RECORD: BreachWatchInfoType
ALTERNATE_PASSWORD: BreachWatchInfoType
GOOD: BWStatus
CHANGED: BWStatus
WEAK: BWStatus
BREACHED: BWStatus
IGNORE: BWStatus

class BreachWatchUpdateRequest(_message.Message):
    __slots__ = ["breachWatchRecordRequest", "encryptedData"]
    BREACHWATCHRECORDREQUEST_FIELD_NUMBER: _ClassVar[int]
    ENCRYPTEDDATA_FIELD_NUMBER: _ClassVar[int]
    breachWatchRecordRequest: _containers.RepeatedCompositeFieldContainer[BreachWatchRecordRequest]
    encryptedData: bytes
    def __init__(self, breachWatchRecordRequest: _Optional[_Iterable[_Union[BreachWatchRecordRequest, _Mapping]]] = ..., encryptedData: _Optional[bytes] = ...) -> None: ...

class BreachWatchRecordRequest(_message.Message):
    __slots__ = ["recordUid", "encryptedData", "breachWatchInfoType", "updateUserWhoScanned"]
    RECORDUID_FIELD_NUMBER: _ClassVar[int]
    ENCRYPTEDDATA_FIELD_NUMBER: _ClassVar[int]
    BREACHWATCHINFOTYPE_FIELD_NUMBER: _ClassVar[int]
    UPDATEUSERWHOSCANNED_FIELD_NUMBER: _ClassVar[int]
    recordUid: bytes
    encryptedData: bytes
    breachWatchInfoType: BreachWatchInfoType
    updateUserWhoScanned: bool
    def __init__(self, recordUid: _Optional[bytes] = ..., encryptedData: _Optional[bytes] = ..., breachWatchInfoType: _Optional[_Union[BreachWatchInfoType, str]] = ..., updateUserWhoScanned: bool = ...) -> None: ...

class BreachWatchData(_message.Message):
    __slots__ = ["passwords", "emails", "domains"]
    PASSWORDS_FIELD_NUMBER: _ClassVar[int]
    EMAILS_FIELD_NUMBER: _ClassVar[int]
    DOMAINS_FIELD_NUMBER: _ClassVar[int]
    passwords: _containers.RepeatedCompositeFieldContainer[BWPassword]
    emails: _containers.RepeatedCompositeFieldContainer[BWPassword]
    domains: _containers.RepeatedCompositeFieldContainer[BWPassword]
    def __init__(self, passwords: _Optional[_Iterable[_Union[BWPassword, _Mapping]]] = ..., emails: _Optional[_Iterable[_Union[BWPassword, _Mapping]]] = ..., domains: _Optional[_Iterable[_Union[BWPassword, _Mapping]]] = ...) -> None: ...

class BWPassword(_message.Message):
    __slots__ = ["value", "resolved", "status", "euid"]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    RESOLVED_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    EUID_FIELD_NUMBER: _ClassVar[int]
    value: str
    resolved: int
    status: BWStatus
    euid: bytes
    def __init__(self, value: _Optional[str] = ..., resolved: _Optional[int] = ..., status: _Optional[_Union[BWStatus, str]] = ..., euid: _Optional[bytes] = ...) -> None: ...



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2022 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import sqlite3
from typing import Callable, Union, Iterable, Tuple

from .types import IEntityStorage, ILinkStorage, IRecordStorage, IUidLink
from . import sqlite_dao


class SqliteEntityStorage(sqlite_dao.SqliteStorage, IEntityStorage):
    def __init__(self, get_connection, schema, owner=None):
        # type: (Callable[[], sqlite3.Connection], sqlite_dao.TableSchema, Union[str, int, None]) -> None

        super(SqliteEntityStorage, self).__init__(get_connection, schema, owner)
        if len(self.schema.primary_key) != 1:
            raise ValueError(f'SqliteEntityStorage: Primary key to have one column.')

    def get_entity(self, uid):
        results = list(self.select_by_filter(self.schema.primary_key, [uid]))
        return results[0] if results else None

    def get_entities(self, pk_values):
        for value in pk_values:
            yield self.get_entity(value)

    def get_all_entities(self):
        for entity in self.select_all():
            yield entity

    def put_entities(self, entities):
        self.put(entities)

    def delete_uids(self, uids):
        self.delete_by_filter(self.schema.primary_key, uids, multiple_criteria=True)


class SqliteLinkStorage(sqlite_dao.SqliteStorage, ILinkStorage):
    def __init__(self, get_connection, schema, owner=None):
        # type: (Callable[[], sqlite3.Connection], sqlite_dao.TableSchema, Union[str, int, None]) -> None

        super(SqliteLinkStorage, self).__init__(get_connection, schema, owner)
        if len(self.schema.primary_key) != 2:
            raise ValueError('SqliteLinkStorage: Primary key to have two columns.')

        object_column = self.schema.primary_key[1]
        object_index_name = None
        if self.schema.indexes:
            for index_name, index_columns in self.schema.indexes.items():
                if index_columns[0].lower() == object_column.lower():
                    object_index_name = index_name
                    break
        if not object_index_name:
            raise ValueError(
                f'SqliteLinkStorage: Object UID column "{object_column}"is not indexed in table "{schema.table_name}".')

    def put_links(self, links):
        self.put(links)

    @staticmethod
    def expand_link_to_tuple(links):
        # type: (Iterable[Union[IUidLink, Tuple[str, str]]]) -> Iterable[Tuple[Union[str, int], Union[str, int]]]
        for link in links:
            if isinstance(link, IUidLink):
                yield link.subject_uid(), link.object_uid()
            elif isinstance(link, (list, tuple)) and len(link) == 2:
                yield link[0], link[1]
            else:
                raise ValueError('Unsupported link type')

    def delete_links(self, links):
        self.delete_by_filter(self.schema.primary_key, SqliteLinkStorage.expand_link_to_tuple(links),
                              multiple_criteria=True)

    def delete_links_for_subjects(self, subject_uids):
        self.delete_by_filter(self.schema.primary_key[0], subject_uids, multiple_criteria=True)

    def delete_links_for_objects(self, object_uids):
        self.delete_by_filter(self.schema.primary_key[1], object_uids, multiple_criteria=True)

    def get_links_for_subject(self, subject_uid):
        for link in self.select_by_filter(self.schema.primary_key[0], subject_uid):
            yield link

    def get_links_for_object(self, object_uid):
        for link in self.select_by_filter(self.schema.primary_key[1], object_uid):
            yield link

    def get_all_links(self):
        for link in self.select_all():
            yield link

    def get_link(self, subject_uid, object_uid):
        for link in self.select_by_filter(self.schema.primary_key, (subject_uid, object_uid)):
            return link
        return None


class SqliteRecordStorage(sqlite_dao.SqliteStorage, IRecordStorage):
    def __init__(self, get_connection, schema, owner):
        # type: (Callable[[], sqlite3.Connection], sqlite_dao.TableSchema, Union[str, int]) -> None
        super(SqliteRecordStorage, self).__init__(get_connection, schema, owner)
        if not schema.owner_column:
            raise ValueError(f'SqliteRecordStorage: Schema \"{schema.table_name}\" should have an owner')
        if schema.primary_key:
            raise ValueError(f'SqliteRecordStorage: Schema \"{schema.table_name}\" should not have primary key')

    def load(self):
        return next(self.select_all(), None)

    def store(self, record):
        self.put([record])

    def delete(self):
        self.delete_all()



from __future__ import annotations
import logging
from . import ConnectionBase
from ..exceptions import DAGConnectionException
from ..crypto import bytes_to_base64
from ..utils import value_to_boolean
import os
import requests
import time

try:  # pragma: no cover
    from ... import crypto, utils, rest_api
except ImportError:  # pragma: no cover
    raise Exception("Please install the keepercommander module to use the Commander connection.")

from typing import Optional, Union, TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from ...params import KeeperParams
    from ...vault import KeeperRecord
    Content = Union[str, bytes, dict]
    QueryValue = Union[list, dict, str, float, int, bool]
    Logger = Union[logging.RootLogger, logging.Logger]


class Connection(ConnectionBase):

    def __init__(self, params: KeeperParams, encrypted_transmission_key: Optional[bytes] = None,
                 encrypted_session_token: Optional[bytes] = None, verify_ssl: bool = True, is_ws: bool = False,
                 logger: Optional[Logger] = None):

        super().__init__(is_device=False, logger=logger)
        self.params = params
        self.verify_ssl = verify_ssl
        self.is_ws = is_ws
        self.encrypted_transmission_key = encrypted_transmission_key if encrypted_transmission_key else None
        self.encrypted_session_token = encrypted_session_token if encrypted_session_token else None

        if self.encrypted_transmission_key is None or self.encrypted_session_token is None:
            self.get_keeper_tokens()

    @staticmethod
    def get_record_uid(record: KeeperRecord) -> str:
        return record.record_uid

    @staticmethod
    def get_key_bytes(record: KeeperRecord) -> bytes:
        return record.record_key

    @property
    def hostname(self) -> str:
        # The host is connect.keepersecurity.com, connect.dev.keepersecurity.com, etc. Append "connect" in front
        # of host used for Commander.
        configured_host = f'connect.{self.params.config.get("server")}'
        
        # In GovCloud environments, the router service is not under the govcloud subdomain
        if 'govcloud.' in configured_host:
            configured_host = configured_host.replace('govcloud.', '') # "connect.govcloud.keepersecurity.com" -> "connect.keepersecurity.com"
            
        return os.environ.get("ROUTER_HOST", configured_host)

    @property
    def dag_server_url(self) -> str:

        # Allow override of the URL. If not set, get the hostname from the config.
        hostname = os.environ.get("KROUTER_URL", self.hostname)
        if hostname.startswith('ws') or hostname.startswith('http'):
            return hostname

        use_ssl = value_to_boolean(os.environ.get("USE_SSL", True))
        if self.is_ws is True:
            prot_pref = 'ws'
        else:
            prot_pref = 'http'
        if use_ssl is True:
            prot_pref += "s"

        return f'{prot_pref}://{hostname}'

    def get_keeper_tokens(self):
        transmission_key = utils.generate_aes_key()
        server_public_key = rest_api.SERVER_PUBLIC_KEYS[self.params.rest_context.server_key_id]

        if self.params.rest_context.server_key_id < 7:
            self.encrypted_transmission_key = crypto.encrypt_rsa(transmission_key, server_public_key)
        else:
            self.encrypted_transmission_key = crypto.encrypt_ec(transmission_key, server_public_key)
        self.encrypted_session_token = crypto.encrypt_aes_v2(
            utils.base64_url_decode(self.params.session_token), transmission_key)

    def rest_call_to_router(self,
                            http_method: str,
                            endpoint: str,
                            agent: str,
                            payload_json: Optional[Union[bytes, str]] = None,
                            retry: int = 5,
                            retry_wait: float = 10,
                            throttle_inc_factor: float = 1.5,
                            timeout: Optional[int] = None) -> str:
        if payload_json is not None and isinstance(payload_json, bytes) is False:
            payload_json = payload_json.encode()

        if endpoint.startswith("/") is False:
            endpoint = "/" + endpoint

        url = self.dag_server_url + endpoint

        attempt = 0
        while True:
            try:
                attempt += 1
                self.logger.debug(f"graph web service call to {url} [{attempt}/{retry}]")
                response = requests.request(
                    method=http_method,
                    url=url,
                    verify=self.verify_ssl,
                    headers={
                        'TransmissionKey': bytes_to_base64(self.encrypted_transmission_key),
                        'Authorization': f'KeeperUser {bytes_to_base64(self.encrypted_session_token)}',
                        'User-Agent': agent
                    },
                    data=payload_json,
                    timeout=timeout
                )
                self.logger.debug(f"response status: {response.status_code}")
                response.raise_for_status()
                return response.text

            except requests.exceptions.HTTPError as http_err:

                msg = http_err.response.reason
                try:
                    content = http_err.response.content.decode()
                    if content is not None and content != "":
                        msg = "; " + content
                except (Exception,):
                    pass

                err_msg = f"{http_err.response.status_code}, {msg}"

                if http_err.response.status_code == 429:
                    attempt -= 1
                    retry_wait *= throttle_inc_factor
                    self.logger.warning("the connection to the graph service is being throttled, "
                                        f"increasing the delay between retry: {retry_wait} seconds.")

            except Exception as err:
                err_msg = str(err)

            self.logger.info(f"call to graph web service had a problem: {err_msg}")
            if attempt >= retry:
                self.logger.error(f"call to graph web service {url}, after {retry} "
                                  f"attempts, failed!: {err_msg}")
                raise DAGConnectionException(f"Call to graph web service {url}, after {retry} "
                                             f"attempts, failed!: {err_msg}")

            self.logger.info(f"will retry call after {retry_wait} seconds.")
            time.sleep(retry_wait)



#!/usr/bin/env python3

"""Provide a class for keeping track of the last server-interaction, and send keep-alives hidden to the user as needed."""

import logging
import time
from . import api, error


class TimeToKeepalive:
    """Keep track of how soon the login timer is to expire, and send a keepalive if we're "too close"."""

    def __init__(self):
        """Initialize."""
        self.update_time_of_last_activity()
        self.server_logout_timer_window = None
        self.lookup_server_logout_window()

    def update_time_of_last_activity(self):
        """Update the time of last activity.  Used on server interaction and class initialization."""
        self.time_of_last_activity = time.time()

    def lookup_server_logout_window(self, params=None):
        """Get the logout_timer value.  If it doesn't exist yet, use None."""
        if self.server_logout_timer_window is not None:
            # We already have it.
            return
        if (
                params is not None and
                params.settings is not None and
                'logout_timer' in params.settings
        ):
            # We should be able to get it.
            self.server_logout_timer_window = float(params.settings['logout_timer']) / 1000.0
            try:
                if isinstance(params.enforcements, dict) and 'longs' in params.enforcements:
                    timeout = next((x['value'] for x in params.enforcements['longs']
                                    if x['key'] == 'logout_timer_desktop'), None)
                    if timeout:
                        timeout = int(timeout) * 60
                        if timeout < self.server_logout_timer_window:
                            self.server_logout_timer_window = timeout
            except Exception as e:
                logging.debug('Error reading logout timeout: %s', e)

    def update(self, params):
        """Update the timer, and possibly issue a keepalive."""
        if not params.session_token:
            return

        current_time = time.time()

        self.lookup_server_logout_window(params)

        if (
                self.server_logout_timer_window is not None and
                (self.server_logout_timer_window / 2) + self.time_of_last_activity < current_time
        ):
            try:
                api.send_keepalive(params)
            except error.KeeperApiError as kae:
                logging.warning(kae.message)


TTK = TimeToKeepalive()



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2025 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

from typing import Tuple
import platform
import logging

from .base import PlatformHandler
from ..utils.constants import PLATFORM_WINDOWS, PLATFORM_DARWIN


class BiometricDetector:
    """Centralized biometric capability detection"""

    def __init__(self):
        self._platform_handlers = self._load_platform_handlers()

    def _load_platform_handlers(self) -> dict:
        """Load available platform handlers"""
        handlers = {}

        try:
            if platform.system() == PLATFORM_WINDOWS:
                from .windows import WindowsHandler
                handlers[PLATFORM_WINDOWS] = WindowsHandler()
        except ImportError:
            logging.debug("Windows platform handler not available")

        try:
            if platform.system() == PLATFORM_DARWIN:
                from .macos import MacOSHandler
                handlers[PLATFORM_DARWIN] = MacOSHandler()
        except ImportError:
            logging.debug("macOS platform handler not available")

        return handlers

    def detect_platform_capabilities(self) -> Tuple[bool, str]:
        """Detect biometric capabilities for current platform"""
        current_platform = platform.system()

        if current_platform not in self._platform_handlers:
            return False, f"Biometric authentication not supported on {current_platform}"

        handler = self._platform_handlers[current_platform]
        return handler.detect_capabilities()

    def get_platform_handler(self) -> PlatformHandler:
        """Get platform handler for current system"""
        current_platform = platform.system()

        if current_platform not in self._platform_handlers:
            raise Exception(f"Biometric authentication not supported on {current_platform}")

        return self._platform_handlers[current_platform] 


from __future__ import annotations
from . import ConnectionBase
from ..utils import value_to_boolean
from ..exceptions import DAGException, DAGConnectionException

from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.serialization import load_der_private_key

try:  # pragma: no cover
    from keeper_secrets_manager_core import utils
    from keeper_secrets_manager_core.configkeys import ConfigKeys
    from keeper_secrets_manager_core.storage import InMemoryKeyValueStorage, KeyValueStorage
    from keeper_secrets_manager_core.utils import url_safe_str_to_bytes, bytes_to_base64, generate_random_bytes
except ImportError:  # pragma: no cover
    raise Exception("Please install the keeper_secrets_manager_core module to use the Ksm connection.")

import logging
import json
import os
import requests
import time

from typing import Union, Optional, TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from keeper_secrets_manager_core.storage import KeyValueStorage
    from keeper_secrets_manager_core.dto.dtos import Record
    KsmConfig = Union[dict, str, KeyValueStorage]
    Content = Union[str, bytes, dict]
    QueryValue = Union[list, dict, str, float, int, bool]
    Logger = Union[logging.RootLogger, logging.Logger]


class Connection(ConnectionBase):

    KEEPER_CLIENT = 'ms16.5.0'

    def __init__(self, config: Union[str, dict, KeyValueStorage], verify_ssl: bool = None,
                 logger: Optional[Logger] = None):

        super().__init__(is_device=True, logger=logger)

        if InMemoryKeyValueStorage.is_base64(config):
            config = utils.base64_to_string(config)
        if isinstance(config, str) is True:
            try:
                config = json.loads(config)
            except json.JSONDecodeError as err:
                raise DAGException(f"The configuration JSON could not be decoded: {err}")

        if isinstance(config, dict) is False and isinstance(config, KeyValueStorage) is False:
            raise DAGException("The configuration is not a dictionary.")

        if verify_ssl is None:
            verify_ssl = value_to_boolean(os.environ.get("VERIFY_SSL", "TRUE"))

        self.config = config
        self.verify_ssl = verify_ssl
        self._signature = None
        self._challenge_str = None

    @staticmethod
    def get_record_uid(record: Record) -> str:
        return record.uid

    @staticmethod
    def get_key_bytes(record: Record) -> bytes:
        return record.record_key_bytes

    def get_config_value(self, key: ConfigKeys) -> str:
        if isinstance(self.config, KeyValueStorage) is True:
            return self.config.get(key)
        else:
            return self.config.get(key.value)

    @property
    def hostname(self) -> str:
        return os.environ.get("ROUTER_HOST", self.get_config_value(ConfigKeys.KEY_HOSTNAME))

    @property
    def client_id(self) -> str:
        return self.get_config_value(ConfigKeys.KEY_CLIENT_ID)

    @property
    def private_key(self) -> str:
        return self.get_config_value(ConfigKeys.KEY_PRIVATE_KEY)

    @property
    def app_key(self) -> str:
        return self.get_config_value(ConfigKeys.KEY_APP_KEY)

    def router_url_from_ksm_config(self) -> str:
        hostname = self.hostname
        # In GovCloud environments, the router service is not under the govcloud subdomain
        if 'govcloud.' in hostname:
            hostname = hostname.replace('govcloud.', '')
        return f'connect.{hostname}'

    def ws_router_url_from_ksm_config(self, is_ws: bool = False) -> str:

        router_host = self.router_url_from_ksm_config()

        kpam_router_ssl_enabled_env = value_to_boolean(os.environ.get("USE_SSL", True))

        if is_ws:
            prot_pref = 'ws'
        else:
            prot_pref = 'http'

        if not kpam_router_ssl_enabled_env:
            return f'{prot_pref}://{router_host}'
        else:
            return f'{prot_pref}s://{router_host}'

    def http_router_url_from_ksm_config_or_env(self) -> str:

        router_host_from_env = os.getenv("KROUTER_URL")
        if router_host_from_env:
            router_http_host = router_host_from_env
        else:
            router_http_host = self.ws_router_url_from_ksm_config()

        return router_http_host.replace('ws', 'http')

    def authenticate(self,
                     agent: str,
                     refresh: bool = False,
                     retry: int = 5,
                     retry_wait: float = 10.0,
                     throttle_inc_factor: float = 1.5,
                     timeout: Optional[int] = None) -> (str, str):

        if self._signature is None or refresh is True:

            self.logger.debug(f"signature is blank or needs to be refresh {refresh}")

            router_http_host = self.http_router_url_from_ksm_config_or_env()
            url = f'{router_http_host}/api/device/get_challenge'

            self._signature = None

            attempt = 0
            while True:
                try:
                    attempt += 1
                    response = requests.get(url,
                                            verify=self.verify_ssl,
                                            timeout=timeout,
                                            headers={
                                                "User-Agent": agent
                                            })
                    response.raise_for_status()

                    self._challenge_str = response.text
                    if self._challenge_str is None or self._challenge_str == "":
                        raise Exception("Challenge text is blank. Cannot authenticate into the DAG web service.")

                    private_key_der_bytes = url_safe_str_to_bytes(self.private_key)
                    client_id_bytes = url_safe_str_to_bytes(self.client_id)

                    self.logger.debug('adding challenge to the signature before connecting to the router')
                    challenge_bytes = url_safe_str_to_bytes(self._challenge_str)
                    client_id_bytes = client_id_bytes + challenge_bytes

                    pk = load_der_private_key(private_key_der_bytes, password=None)
                    sig = pk.sign(client_id_bytes, ec.ECDSA(hashes.SHA256()))

                    self._signature = bytes_to_base64(sig)
                    break

                except requests.exceptions.HTTPError as http_err:

                    msg = http_err.response.reason
                    try:
                        content = http_err.response.content.decode()
                        if content is not None and content != "":
                            msg = "; " + content
                    except (Exception,):
                        pass

                    err_msg = f"{http_err.response.status_code}, {msg}"

                    if http_err.response.status_code == 429:
                        retry_wait *= throttle_inc_factor
                        attempt -= 1
                        self.logger.warning(
                            "the connection to the graph service, for authentication, is being throttled; "
                            f"increasing delay between retry: {retry_wait} seconds")

                except Exception as err:
                    err_msg = str(err)

                self.logger.info(f"call to challenge had a problem: {err_msg}.")
                if attempt >= retry:
                    raise DAGConnectionException(f"Call to challenge {url}, after {retry} "
                                                 f"attempts, failed!: {err_msg}")

                self.logger.info(f"will retry call after {retry_wait} seconds.")
                time.sleep(retry_wait)

        return self._signature, self._challenge_str

    def rest_call_to_router(self,
                            http_method: str,
                            endpoint: str,
                            agent: str,
                            payload_json: Optional[Union[bytes, str]] = None,
                            retry: int = 5,
                            retry_wait: float = 10.0,
                            throttle_inc_factor: float = 1.5,
                            timeout: Optional[int] = None) -> str:

        # If the timeout is set to 0, set to the default which is None.
        if timeout == 0:
            timeout = None

        router_host = self.http_router_url_from_ksm_config_or_env()
        url = router_host + endpoint

        refresh = False
        attempt = 0
        while True:

            attempt += 1

            # Keep authenticate outside the call router try.
            # This is to prevent too many retries.
            # For example, 3 retry of the auth, 3 retry of the request, will be 9 retries.
            signature, challenge_str = self.authenticate(refresh=refresh, agent=agent)
            headers = {
                "Signature": signature,
                "ClientVersion": Connection.KEEPER_CLIENT,
                "Authorization": f'KeeperDevice {self.client_id}',
                "Challenge": challenge_str,
                "User-Agent": agent
            }
            self.logger.debug(f'connecting with headers: {headers}')

            try:
                self.logger.debug(f"DAG web service call to {url} [{attempt}/{retry}]")
                response = requests.request(
                    method=http_method,
                    url=url,
                    data=payload_json if payload_json else None,
                    verify=self.verify_ssl,
                    timeout=timeout,
                    headers=headers,
                )

                self.logger.debug(f"response status: {response.status_code}")

                # If we get a 401 Unauthorized, and we have not yet refreshed,
                #  refresh the signature.
                if response.status_code == 401 and refresh is False:
                    self.logger.debug("rest call was Unauthorized")

                    # The attempt didn't count.
                    # We get one refresh, then it becomes an exception.
                    refresh = True
                    attempt -= 1
                    continue

                response.raise_for_status()
                return response.text

            # Handle errors outside of requests
            except requests.exceptions.HTTPError as http_err:

                err_msg = f"{http_err.response.status_code}, {http_err.response.reason}, {http_err.response.content}"
                content = http_err.response.reason

                if http_err.response.status_code == 429:
                    retry_wait *= throttle_inc_factor
                    attempt -= 1
                    self.logger.warning("the connection to the graph service is being throttled, "
                                        f"increasing the delay between retry: {retry_wait} seconds.")

            except Exception as err:
                err_msg = str(err)
                content = None

            self.logger.info(f"call to graph web service had a problem: {err_msg}, {content}")
            if attempt >= retry:
                self.logger.info(f"payload: {payload_json}")
                raise DAGConnectionException(f"Call to graph web service {url}, after {retry} "
                                             f"attempts, failed!: {err_msg}: {content} : {payload_json}")

            self.logger.info(f"will retry call after {retry_wait} seconds.")
            time.sleep(retry_wait)



from __future__ import annotations
import argparse
import logging
import json
from . import PAMGatewayActionDiscoverCommandBase, GatewayContext
from .job_status import PAMGatewayActionDiscoverJobStatusCommand
from ..pam.router_helper import router_send_action_to_gateway, print_router_response, router_get_connected_gateways
from ..pam.user_facade import PamUserRecordFacade
from ..pam.pam_dto import GatewayActionDiscoverJobStartInputs, GatewayActionDiscoverJobStart, GatewayAction
from ... import vault_extensions
from ... import vault
from ...proto import pam_pb2
from ...display import bcolors
from ...discovery_common.jobs import Jobs
from ...discovery_common.types import CredentialBase
from typing import List, TYPE_CHECKING

if TYPE_CHECKING:
    from ...params import KeeperParams


class PAMGatewayActionDiscoverJobStartCommand(PAMGatewayActionDiscoverCommandBase):
    parser = argparse.ArgumentParser(prog='pam-action-discover-start')
    parser.add_argument('--gateway', '-g', required=True, dest='gateway', action='store',
                        help='Gateway name of UID.')
    parser.add_argument('--resource', '-r', required=False, dest='resource_uid', action='store',
                        help='UID of the resource record. Set to discover specific resource.')
    parser.add_argument('--lang', required=False, dest='language', action='store', default="en_US",
                        help='Language')
    parser.add_argument('--include-machine-dir-users', required=False, dest='include_machine_dir_users',
                        action='store_false', default=True, help='Include directory users found on the machine.')
    parser.add_argument('--inc-azure-aadds', required=False, dest='include_azure_aadds',
                        action='store_true', help='Include Azure Active Directory Domain Service.')
    parser.add_argument('--skip-rules', required=False, dest='skip_rules',
                        action='store_true', help='Skip running the rule engine.')
    parser.add_argument('--skip-machines', required=False, dest='skip_machines',
                        action='store_true', help='Skip discovering machines.')
    parser.add_argument('--skip-databases', required=False, dest='skip_databases',
                        action='store_true', help='Skip discovering databases.')
    parser.add_argument('--skip-directories', required=False, dest='skip_directories',
                        action='store_true', help='Skip discovering directories.')
    parser.add_argument('--skip-cloud-users', required=False, dest='skip_cloud_users',
                        action='store_true', help='Skip discovering cloud users.')
    parser.add_argument('--cred', required=False, dest='credentials',
                        action='append', help='List resource credentials.')
    parser.add_argument('--cred-file', required=False, dest='credential_file',
                        action='store', help='A JSON file containing list of credentials.')

    def get_parser(self):
        return PAMGatewayActionDiscoverJobStartCommand.parser

    @staticmethod
    def make_protobuf_user_map(params: KeeperParams, gateway_context: GatewayContext) -> List[dict]:
        """
        Make a user map for PAM Users.

        The map is used to find existing records.
        Since KSM cannot read the rotation settings using protobuf,
          it cannot match a vault record to a discovered users.
        This map will map a login/DN and parent UID to a record UID.
        """

        user_map = []
        for record in vault_extensions.find_records(params, record_type="pamUser"):
            user_record = vault.KeeperRecord.load(params, record.record_uid)
            user_facade = PamUserRecordFacade()
            user_facade.record = user_record

            info = params.record_rotation_cache.get(user_record.record_uid)
            if info is None:
                continue

            # Make sure this user is part of this gateway.
            if info.get("configuration_uid") != gateway_context.configuration_uid:
                continue

            # If the user Admin Cred Record (i.e., parent) is blank, skip the mapping item
            # This will be a UID string, not 16 bytes.
            if info.get("resource_uid") is None or info.get("resource_uid") == "":
                continue

            user_map.append({
                "user": user_facade.login if user_facade.login != "" else None,
                "dn": user_facade.distinguishedName if user_facade.distinguishedName != "" else None,
                "record_uid": user_record.record_uid,
                "parent_record_uid": info.get("resource_uid")
            })

        logging.debug(f"found {len(user_map)} user map items")

        return user_map

    def execute(self, params, **kwargs):

        if not hasattr(params, 'pam_controllers'):
            router_get_connected_gateways(params)

        # Load the configuration record and get the gateway_uid from the facade.
        gateway = kwargs.get('gateway')

        gateway_context = GatewayContext.from_gateway(params, gateway)
        if gateway_context is None:
            print(f"{bcolors.FAIL}Could not find the gateway configuration for {gateway}.")
            return

        jobs = Jobs(record=gateway_context.configuration, params=params)
        current_job_item = jobs.current_job
        removed_prior_job = None
        if current_job_item is not None:
            if current_job_item.is_running is True:
                print("")
                print(f"{bcolors.FAIL}A discovery job is currently running. "
                      f"Cannot start another until it is finished.{bcolors.ENDC}")
                print(f"To check the status, use the command "
                      f"'{bcolors.OKGREEN}pam action discover status{bcolors.ENDC}'.")
                print(f"To stop and remove the current job, use the command "
                      f"'{bcolors.OKGREEN}pam action discover remove -j {current_job_item.job_id}'.")
                return

            print(f"{bcolors.FAIL}An active discovery job exists for this gateway.{bcolors.ENDC}")
            print("")
            status = PAMGatewayActionDiscoverJobStatusCommand()
            status.execute(params=params)
            print("")

            yn = input("Do you wish to remove the active discovery job and run a new one [Y/N]> ").lower()
            while True:
                if yn[0] == "y":
                    jobs.cancel(current_job_item.job_id)
                    removed_prior_job = current_job_item.job_id
                    break
                elif yn[0] == "n":
                    print(f"{bcolors.FAIL}Not starting a discovery job.{bcolors.ENDC}")
                    return

        # Get the credentials passed in via the command line
        credentials = []
        creds = kwargs.get('credentials')
        if creds is not None:
            for cred in creds:
                parts = cred.split("|")
                c = CredentialBase()
                for item in parts:
                    kv = item.split("=")
                    if len(kv) != 2:
                        print(f"{bcolors.FAIL}A '--cred' is invalid. It does not have a value.{bcolors.ENDC}")
                        return
                    if hasattr(c, kv[0]) is False:
                        print(f"{bcolors.FAIL}A '--cred' is invalid. The key '{kv[0]}' is invalid.{bcolors.ENDC}")
                        return
                    if hasattr(c, kv[1]) == "":
                        print(f"{bcolors.FAIL}A '--cred' is invalid. The value is blank.{bcolors.ENDC}")
                        return
                    setattr(c, kv[0], kv[1])
                credentials.append(c.model_dump())

        # Get the credentials passed in via a credential file.
        credential_files = kwargs.get('credential_file')
        if credential_files is not None:
            with open(credential_files, "r") as fh:
                try:
                    creds = json.load(fh)
                except FileNotFoundError:
                    print(f"{bcolors.FAIL}Could not find the file {credential_files}{bcolors.ENDC}")
                    return
                except json.JSONDecoder:
                    print(f"{bcolors.FAIL}The file {credential_files} is not valid JSON.{bcolors.ENDC}")
                    return
                except Exception as err:
                    print(f"{bcolors.FAIL}The JSON file {credential_files} could not be imported: {err}{bcolors.ENDC}")
                    return

                if isinstance(creds, list) is False:
                    print(f"{bcolors.FAIL}Credential file is invalid. Structure is not an array.{bcolors.ENDC}")
                    return
                num = 1
                for obj in creds:
                    c = CredentialBase()
                    for key in obj:
                        if hasattr(c, key) is False:
                            print(f"{bcolors.FAIL}Object {num} has the invalid key {key}.{bcolors.ENDC}")
                            return
                        setattr(c, key, obj[key])
                    credentials.append(c.model_dump())

        action_inputs = GatewayActionDiscoverJobStartInputs(
            configuration_uid=gateway_context.configuration_uid,
            resource_uid=kwargs.get('resource_uid'),
            user_map=gateway_context.encrypt(
                self.make_protobuf_user_map(
                    params=params,
                    gateway_context=gateway_context
                )
            ),

            shared_folder_uid=gateway_context.default_shared_folder_uid,
            languages=[kwargs.get('language')],

            # Settings
            include_machine_dir_users=kwargs.get('include_machine_dir_users', True),
            include_azure_aadds=kwargs.get('include_azure_aadds', False),
            skip_rules=kwargs.get('skip_rules', False),
            skip_machines=kwargs.get('skip_machines', False),
            skip_databases=kwargs.get('skip_databases', False),
            skip_directories=kwargs.get('skip_directories', False),
            skip_cloud_users=kwargs.get('skip_cloud_users', False),
            credentials=credentials
        )

        conversation_id = GatewayAction.generate_conversation_id()
        router_response = router_send_action_to_gateway(
            params=params,
            gateway_action=GatewayActionDiscoverJobStart(
                inputs=action_inputs,
                conversation_id=conversation_id),
            message_type=pam_pb2.CMT_DISCOVERY,
            is_streaming=False,
            destination_gateway_uid_str=gateway_context.gateway_uid
        )

        data = self.get_response_data(router_response)
        if data is None:
            print(f"{bcolors.FAIL}The router returned a failure.{bcolors.ENDC}")
            return

        if "has been queued" in data.get("Response", ""):

            if removed_prior_job is None:
                print("The discovery job is currently running.")
            else:
                print(f"Active discovery job {removed_prior_job} has been removed and new discovery job is running.")
            print(f"To check the status, use the command '{bcolors.OKGREEN}pam action discover status{bcolors.ENDC}'.")
            print(f"To stop and remove the current job, use the command "
                  f"'{bcolors.OKGREEN}pam action discover remove -j <Job ID>'.")
        else:
            print_router_response(router_response, "job_info", conversation_id, gateway_uid=gateway_context.gateway_uid)



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2018 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#
import logging
from typing import Optional, Tuple, Dict, Iterable, List, Set, Union

from .params import KeeperParams


def get_folder_path(params, folder_uid, delimiter='/'):
    uid = folder_uid
    path = ''
    while uid in params.folder_cache:
        f = params.folder_cache[uid]
        name = f.name
        name = name.replace(delimiter, 2*delimiter)
        if len(path) > 0:
            path = name + delimiter + path
        else:
            path = name
        uid = f.parent_uid
    return path


def find_folders(params, record_uid):   # type: (KeeperParams, str) -> Iterable[str]
    for fuid in params.subfolder_record_cache:
        if record_uid in params.subfolder_record_cache[fuid]:
            if fuid:
                yield fuid


def find_all_folders(params, record_uid):   # type: (KeeperParams, str) -> Iterable[BaseFolderNode]
    for fuid in params.subfolder_record_cache:
        if record_uid in params.subfolder_record_cache[fuid]:
            if fuid:
                if fuid in params.folder_cache:
                    yield params.folder_cache[fuid]
            else:
                yield params.root_folder


def find_parent_top_folder(params, record_uid):

    """
    Find all top Shared Folders that will contain this record.
    Record can be in more than two folders by having a "link" which
    will present a record as a record with the same UID in more than
    one folder.
    """
    contained_folder_uids = []

    # Get all folders that might contain the given record
    for fuid in params.subfolder_record_cache:
        if fuid:    # record is in root folder
            if record_uid in params.subfolder_record_cache[fuid]:
                contained_folder_uids.append(fuid)

    shared_folders_containing_record = []

    # if records that belong to folder are found then go
    # through each one of them to get the share folder
    for cfuid in contained_folder_uids:
        if cfuid:
            folder = params.folder_cache[cfuid]

            # folder is already a shared folder
            if isinstance(folder, SharedFolderNode):
                shared_folders_containing_record.append(folder)
            # folder is a sub-folder, let's get its pared shared folder
            elif isinstance(folder, SharedFolderFolderNode):
                shared_folder_uid = folder.shared_folder_uid
                shared_folder = params.folder_cache[shared_folder_uid]
                shared_folders_containing_record.append(shared_folder)
            else:
                logging.debug("Folder UID={} is not a shared folder".format(cfuid))

    return shared_folders_containing_record


def contained_folders(params, folders, component):
    # type: (KeeperParams, List[Optional[BaseFolderNode]], str) -> List[Optional[BaseFolderNode]]
    """Return list of folders (empty if component not present) containing component within parent folder 'folder'"""
    get_folder_by_id = lambda uid: params.folder_cache.get(uid)
    get_folder_ids = lambda: params.folder_cache.keys()
    result = folders if component in ('.', '') \
        else [(get_folder_by_id(f.parent_uid) if f.parent_uid else params.root_folder) for f in folders] if component == '..' \
        else [get_folder_by_id(component)] if component in get_folder_ids() \
        else [get_folder_by_id(uid) for f in folders for uid in f.subfolders if get_folder_by_id(uid).name == component]
    return result


def lookup_path(params, folder, components):
    # type: (KeeperParams, BaseFolderNode, List[Optional[str]]) -> Tuple[int, List[Optional[BaseFolderNode]]]
    """
    Lookup path of components within the folder cache.

    Get all the folders starting from the left end of component, plus the index of the first component that isn't
    present in the folder cache.
    """
    remainder = 0
    folders = [folder]
    for index, component in enumerate(components):
        child_folders = contained_folders(params, folders, component)
        if not child_folders:
            break
        folders = child_folders
        remainder = index + 1
    return remainder, folders


def is_abs_path(path_string):
    """Return True iff path_string is an absolute path."""
    return path_string.startswith('/') and not path_string.startswith('//')


def path_split(params, folder, path_string):
    """Split a path into directories with two replaces and a split."""
    if is_abs_path(path_string):
        folder = params.root_folder
        path_string = path_string[1:]

    components = [s.replace('\0', '/') for s in path_string.replace('//', '\0').split('/')]
    return folder, components


def try_resolve_path(params, path, find_all_matches=False):
    # type: (KeeperParams, str, bool) -> Tuple[Union[List[BaseFolderNode], BaseFolderNode, None], Optional[str]]
    """
    Look up the final keepercommander.subfolder.UserFolderNode and name of the final component(s).
    Set find_all_matches = True to get a list of folders and component path

    If a record, the final component is the record.
    If existent folder(s), the final component is ''.
    If a non-existent folder, the final component is the folders, joined with /, that do not (yet) exist..
    """
    if type(path) is not str:
        path = ''

    folder = (
        params.folder_cache[params.current_folder]
        if params.current_folder in params.folder_cache
        else params.root_folder
    )

    folder, components = path_split(params, folder, path)

    remainder, folders = lookup_path(params, folder, components)

    tail = components[remainder:]

    path = '/'.join(component.replace('/', '//') for component in tail)

    # Return a 2-tuple of BaseFolderNode (or List[BaseFolderNode] if find_all_matches set to True), str
    # The first is the folder/s containing the second, or the folder of the last component if the second is ''.
    # The second is the final component of the path we're passed as an argument to this function. It could be a record, or
    # a not-yet-existent directory.
    return (folders, path) if find_all_matches \
        else (next(iter(folders)) if folders else None, path)


def get_folder_uids(params, name):  # type: (KeeperParams, str or None) -> Set[Optional[str]]
    uids = set()
    if name in params.folder_cache or name == '':
        uids.add(name)
    else:
        rs = try_resolve_path(params, name, find_all_matches=True)
        if rs is not None:
            folders, pattern = rs
            if len(pattern) == 0:
                uids.update([folder.uid or '' for folder in folders])
    return uids


def get_contained_folder_uids(params, name, children_only=True):
    def on_folder(f):
        f_uid = f.uid or ''
        parent_uid = f.parent_uid or ''
        if f_uid and f_uid not in root_folder_uids and (not children_only or parent_uid in root_folder_uids):
            folder_uids.add(f_uid)

    folder_uids = set()
    root_folder_uids = get_folder_uids(params, name)
    from .commands.base import FolderMixin
    for uid in get_folder_uids(params, name):
        FolderMixin.traverse_folder_tree(params, uid, on_folder)

    return folder_uids


def get_contained_record_uids(params, name, children_only=True):
    # type: (KeeperParams, str, bool) -> Dict[str, Iterable[str]]
    from .commands.base import FolderMixin
    recs_by_folder = dict()
    root_folder_uids = get_folder_uids(params, name)

    def add_child_recs(f_uid):
        child_recs = params.subfolder_record_cache.get(f_uid, set())
        recs_by_folder.update({f_uid: child_recs})

    def on_folder(f):  # type: (BaseFolderNode) -> None
        f_uid = f.uid or ''
        if not children_only or f_uid in root_folder_uids:
            add_child_recs(f_uid)

    for uid in root_folder_uids:
        FolderMixin.traverse_folder_tree(params, uid, on_folder)

    return recs_by_folder


class BaseFolderNode:
    RootFolderType = '/'
    UserFolderType = 'user_folder'
    SharedFolderType = 'shared_folder'
    SharedFolderFolderType = 'shared_folder_folder'

    """ Folder Common Fields"""
    def __init__(self, type):
        self.type = type
        self.uid = None
        self.parent_uid = None
        self.name = None     # type: Optional[str]
        self.color = None    # type: Optional[str]
        self.subfolders = []

    def get_folder_type(self):
        if self.type == BaseFolderNode.RootFolderType:
            return 'Root'
        elif self.type == BaseFolderNode.UserFolderType:
            return 'Personal Folder'
        elif self.type == BaseFolderNode.SharedFolderType:
            return 'Shared Folder'
        elif self.type == BaseFolderNode.SharedFolderFolderType:
            return 'Subfolder in Shared Folder'
        return ''

    def __repr__(self):
        return 'BaseFolderNode(type={}, uid={}, parent_uid={}, name={}, subfolders={})'.format(
            self.type,
            self.uid,
            self.parent_uid,
            self.name,
            self.subfolders,
        )

    def display(self):
        print('')
        print('{0:>20s}: {1:<20s}'.format('Folder UID', self.uid))
        print('{0:>20s}: {1:<20s}'.format('Folder Type', self.get_folder_type()))
        print('{0:>20s}: {1}'.format('Name', self.name))
        if self.parent_uid:
            print('{0:>20s}: {1:<20s}'.format('Parent Folder UID', self.parent_uid))
        if isinstance(self, SharedFolderFolderNode):
            print('{0:>20s}: {1:<20s}'.format('Shared Folder UID', self.shared_folder_uid))


class UserFolderNode(BaseFolderNode):
    def __init__(self):
        BaseFolderNode.__init__(self, BaseFolderNode.UserFolderType)


class SharedFolderFolderNode(BaseFolderNode):
    def __init__(self):
        BaseFolderNode.__init__(self, BaseFolderNode.SharedFolderFolderType)
        self.shared_folder_uid = None


class SharedFolderNode(BaseFolderNode):
    def __init__(self):
        BaseFolderNode.__init__(self, BaseFolderNode.SharedFolderType)

    @property
    def shared_folder_uid(self):
        return self.uid

    @shared_folder_uid.setter
    def shared_folder_uid(self, val):
        self.uid = val


class RootFolderNode(BaseFolderNode):
    def __init__(self):
        BaseFolderNode.__init__(self, BaseFolderNode.RootFolderType)
        self.name = 'My Vault'
